/*

Copyright (c) 2002-2005, Yauheni Akhotnikau
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

- The name of the author may not be used to endorse or promote products derived
from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.

*/

/*!
	\since v.1.4.0
	\file
	\brief     / 
		      .
*/

#include <oess_1/defs/h/ex.hpp>

#include <oess_1/io/h/bin_data_size.hpp>

#include <oess_1/db/storage/h/errno.hpp>

#include <oess_1/db/storage/h/page_index_calculator.hpp>

#include <oess_1/db/storage/impl/h/segment_header_item.hpp>

namespace oess_1 {

namespace db {

namespace storage {

//!     ,   .
/*!  31- , ..    POSIX 
	  31-  */
const oess_1::uint_t max_file_size = 0x7fffffffu;

//!    .
const oess_1::uint_t max_page_size = 64 * 1024;

//!    .
/*!         
	 . */
const oess_1::uint_t min_page_size = 
		2 * impl::segment_header_item_t::image_size;

//
// page_index_calculator_t
//

page_index_calculator_t::page_index_calculator_t(
	oess_1::uint_t page_size,
	oess_1::uint_t chain_link_size )
	:	m_page_size( page_size )
	,	m_chain_link_size( chain_link_size )
	,	m_page_chain_links( 0 )
	,	m_areas( 0 )
	,	m_area_segments( 0 )
	,	m_last_area_segments( 0 )
	,	m_area_pages( 0 )
	,	m_segment_pages( 0 )
	{
		//         
		//    .
		if( page_size < min_page_size || page_size > max_page_size )
			OESS_THROW_LOGIC(
					oess_1::db::storage::err::c_invalid_sizes,
					"invalid page_size: " << chain_link_size
					<< "; must be >= " << min_page_size
					<< " && <= " << max_page_size )

		//        
		//  .
		if( chain_link_size < impl::segment_header_item_t::image_size )
			OESS_THROW_LOGIC(
					oess_1::db::storage::err::c_invalid_sizes,
					"invalid chain_link_size: " << chain_link_size
					<< "; must be at least: "
					<< impl::segment_header_item_t::image_size )

		//         .
		if( page_size % chain_link_size )
			OESS_THROW_LOGIC(
					oess_1::db::storage::err::c_invalid_sizes,
					"page_size % chain_link_size must be 0; "
					"page_size: " << page_size
					<< ", chain_link_size: " << chain_link_size )

		//         
		//  .
		if( 2 > ( page_size / impl::segment_header_item_t::image_size ) )
			OESS_THROW_LOGIC(
					oess_1::db::storage::err::c_invalid_sizes,
					"invalid page_size: " << page_size
					<< "; must be at least: "
					<< 2 * impl::segment_header_item_t::image_size )

		m_page_chain_links = page_size / chain_link_size;

		//      .
		oess_1::uint_t available_pages = max_file_size / page_size - 1;

		m_segment_pages = page_size /
				impl::segment_header_item_t::image_size /
				m_page_chain_links;

		//       .
		m_area_segments = page_size / oess_1::io::bin_data_size_t<
						segment_ordinal_t >::image_size
				- 1;

		//       .
		m_area_pages = m_area_segments * m_segment_pages + 1;

		//   .
		m_areas = available_pages / m_area_pages;
		//   ,     
		//   1.
		oess_1::uint_t remaining_pages = available_pages % m_area_pages;
		if( remaining_pages > ( m_segment_pages + 1 ) )
			{
				++m_areas;
				m_last_area_segments =
						( remaining_pages - 1 ) / m_segment_pages;
			}
		else
			m_last_area_segments = m_area_segments;

		//      .
		oess_1::uint_t common_header_items =
				page_size / oess_1::io::bin_data_size_t<
						area_ordinal_t >::image_size
				- 1;
		if( m_areas > common_header_items )
			{
				// ..     ,  
				//   ,   
				//    .
				m_areas = common_header_items;
				m_last_area_segments = m_area_segments;
			}
	}

oess_1::uint_t
page_index_calculator_t::common_header_page() const
	{
		//   .
		return 0;
	}

area_ordinal_t
page_index_calculator_t::area_count() const
	{
		return m_areas;
	}

segment_ordinal_t
page_index_calculator_t::area_segment_count(
	const area_ordinal_t & area ) const
	{
		if( area + 1 != m_areas )
			//   .
			return m_area_segments;
		else
			//  , ,   .
			return m_last_area_segments;
	}

oess_1::uint_t
page_index_calculator_t::area_header_page(
	const area_ordinal_t & area ) const
	{
		//   --   .
		return 1 + area * m_area_pages;
	}

oess_1::uint_t
page_index_calculator_t::segment_header_page(
	const area_ordinal_t & area,
	const segment_ordinal_t & segment ) const
	{
		//   --   .
		return area_header_page( area ) +
				m_segment_pages * segment + 1;
	}

oess_1::uint_t
page_index_calculator_t::segment_page_count() const
	{
		return m_segment_pages;
	}

oess_1::uint_t
page_index_calculator_t::chain_link_on_page_count() const
	{
		return m_page_chain_links;
	}

chain_id_t
page_index_calculator_t::first_chain_link_on_page(
	oess_1::uint_t page_ordinal ) const
	{
		return page_ordinal * m_page_chain_links;
	}

area_ordinal_t
page_index_calculator_t::chain_link_area(
	const chain_id_t & block ) const
	{
		oess_1::uint_t page = chain_link_page( block );
		if( !page )
			OESS_THROW_LOGIC(
					oess_1::db::storage::err::c_invalid_chain_link_ordinal,
					block << ": cannot resides on common header page" )

		return ( page - 1 ) / m_area_pages;
	}

segment_ordinal_t
page_index_calculator_t::chain_link_segment(
	const chain_id_t & block ) const
	{
		oess_1::uint_t page = chain_link_page( block );
		if( !page )
			OESS_THROW_LOGIC(
					oess_1::db::storage::err::c_invalid_chain_link_ordinal,
					block << ": cannot resides on common header page" )

		oess_1::uint_t inside_area = ( page - 1 ) % m_area_pages;
		if( !inside_area )
			OESS_THROW_LOGIC(
					oess_1::db::storage::err::c_invalid_chain_link_ordinal,
					block << ": cannot resides on area header page" )

		return ( inside_area - 1 ) / m_segment_pages;
	}

oess_1::uint_t
page_index_calculator_t::chain_link_page(
	const chain_id_t & block ) const
	{
		return block / m_page_chain_links;
	}

oess_1::uint_t
page_index_calculator_t::chain_link_page_offset(
	const chain_id_t & block ) const
	{
		return ( block % m_page_chain_links ) * m_chain_link_size;
	}

chain_id_t
page_index_calculator_t::first_chain_id() const
	{
		return first_chain_link_on_page(
				segment_header_page( 0, 0 ) + 1 );
	}

} /* namespace storage */

} /* namespace db */

} /* namespace oess_1 */

