/*

Copyright (c) 2002-2005, Yauheni Akhotnikau
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

- The name of the author may not be used to endorse or promote products derived
from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.

*/

/*!
	\file
	\since v.1.4.0
	\brief    trace-, 
		 .
*/

#include <ace/Dirent.h>
#include <ace/Dirent_Selector.h>

#include <pcre++.h>

#include <cpp_util_2/h/lexcast.hpp>
#include <cpp_util_2/lexcasts/h/util.hpp>

#include <oess_1/defs/h/ex.hpp>

#include <oess_1/db/storage/h/errno.hpp>

#include <oess_1/db/storage/trx/h/find_all_trace_files.hpp>

namespace oess_1 {

namespace db {

namespace storage {

namespace trx {

//
// split_physic_name
//

/*!
	         .
*/
static void
split_physic_name(
	//!  , , ,   .
	const std::string & physic_name,
	//!    .
	std::string & path_name,
	//!    .
	std::string & db_name )
	{
		//  ,       
		//  ,    .
		try
			{
				pcrepp::Pcre name_sep_expr(
						"^(.*[\\\\/:]){0,1}([^\\\\/:]+$)" );
				if( !name_sep_expr.search( physic_name ) )
					OESS_THROW_PHYSIC(
							oess_1::db::storage::err::c_unexpected,
							physic_name 
							<< ": unable to split to <path> and <name>; "
									"Pcre::search failed!" )

				//    ,   .
				path_name = name_sep_expr[ 0 ];
				db_name = name_sep_expr[ 1 ];
			}
		catch( const std::exception & x )
			{
				OESS_THROW_PHYSIC(
						oess_1::db::storage::err::c_unexpected,
						physic_name 
						<< ": unable to split to <path> and <name>; "
								"exception caught: " << x.what() )
			}
	}

//
// process_trace_files
//

/*!
	   trace-  snapshot-    
	     trace-  snapshot-.
*/
static void
process_found_files(
	//!  .
	ACE_Dirent_Selector & selector,
	//!   ,     .
	const std::string & physic_name,
	//! ,     .
	const std::string & extension,
	//!   trace-.
	trace_file_number_set_t & numbers )
	{
		try
			{
				//  ,   
				//      .
				pcrepp::Pcre file_regex(
						"^" +
						physic_name + ".([[:xdigit:]]{8})" + extension +
						"$" );
   
				for( int i = 0; i != selector.length(); ++i )
					{
						if( file_regex.search( selector[ i ]->d_name ) )
							{
								trace_file_number_t n;
								std::sscanf(
										file_regex[ 0 ].c_str(), "%x", &n );
								numbers.insert( n );
							}
					}
			}
		catch( const std::exception & x )
			{
				OESS_THROW_PHYSIC(
						oess_1::db::storage::err::c_unexpected,
						physic_name 
						<< ": unable to filter found file names; "
								"extension: " << extension
						<< "exception caught: " << x.what() )
			}
	}

//
// find_all_specified_files
//

/*!
	      ,  
	 .
*/
trace_file_number_set_t
find_all_specified_files(
	const std::string & physic_name,
	const std::string & extension )
	{
		//   ,      
		//   .
		std::string path_name, db_name;
		split_physic_name( physic_name, path_name, db_name );

		//        ,
		//       trace-.
		ACE_Dirent_Selector selector;
		if( -1 == selector.open( path_name.c_str() ) )
			OESS_THROW_PHYSIC(
					oess_1::db::storage::err::c_trx_file_search_failed,
					"unable to search files in: " << path_name
					<< "; extension: " << extension )

		trace_file_number_set_t numbers;
		process_found_files( selector, db_name, extension, numbers );

		return numbers;
	}

//
// find_all_trace_files
//

trace_file_number_set_t
find_all_trace_files(
	const std::string & physic_name )
	{
		return find_all_specified_files( physic_name, trace_file_extension() );
	}

//
// find_all_snapshot_files
//

trace_file_number_set_t
find_all_snapshot_files(
	const std::string & physic_name )
	{
		return find_all_specified_files(
				physic_name, snapshot_file_extension() );
	}

} /* namespace trx */

} /* namespace storage */

} /* namespace db */

} /* namespace oess */

