/*

Copyright (c) 2002-2005, Yauheni Akhotnikau
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

- The name of the author may not be used to endorse or promote products derived
from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.

*/

/*!
	\since v.1.4.0
	\brief  %std_file_committer_t.
*/
#if !defined( _OESS_1__DB__STORAGE__TRX__STD_FILE_COMMITTER_HPP_ )
#define _OESS_1__DB__STORAGE__TRX__STD_FILE_COMMITTER_HPP_

#include <vector>

#include <oess_1/file/h/file.hpp>

#include <oess_1/db/storage/trx/h/types.hpp>

namespace oess_1 {

namespace db {

namespace storage {

namespace trx {

//
// std_file_committer_t
//

/*!
	\since v.1.4.0
	\brief  %committer_t,   
		     .

	  detect_nonactual_traces()    
	trace-  trace-,     snapshot.
	  snapshot- ,    
	 (..  trace-  ).

	  initialize()     
	  .    snapshot, 
	      
	snapshot-     snapshot-  
	 .

	  finalize()     
	  .

	  finalize()   (..  
	),      .

	 is_insufficient_trace_info_allowed()  true,  
	    snapshot-.
*/
class	OESS_1__DB__TYPE	std_file_committer_t
	:	public committer_t
	{
	public :
		//!  .
		std_file_committer_t(
			//!    .
			const std::string & physic_name );
		virtual ~std_file_committer_t();

	//! \name   .
	//! \{
		virtual void
		detect_nonactual_traces(
			trace_file_number_set_t & traces );

		/*!
			\note  total_page_count    !
		*/
		virtual void
		initialize(
			oess_1::uint_t total_page_count,
			oess_1::uint_t page_size,
			oess_1::uint_t block_size );

		virtual void
		finalize();

		virtual void
		store_block(
			oess_1::uint_t ordinal,
			oess_1::uint_t block_index,
			const std::string & block_value );

		virtual bool
		is_insufficient_trace_info_allowed() const;
	//! \}

		//!        .
		const std::string &
		physic_name() const;

		//!   ,    
		//!      
		//!  .
		void
		set_result_file_name(
			const std::string & file_name );

	private :
		//!    .
		std::string	m_physic_name;

		//!  snapshot-,     .
		std::auto_ptr< snapshot_collection_t >	m_snapshots;

		//!     .
		oess_1::file::file_t	m_file;

		//!  ,       .
		/*!
			  initialize(),     .

			  finalize(),     . 
		*/
		bool	m_is_file_exists;

		//!       .
		/*!
			  initialize().
		*/
		oess_1::uint_t	m_page_size;

		//!   /   store_block.
		std::vector< oess_1::char_t >	m_io_buf;

		//!  ,      
		//!     .
		/*!
			       .
		*/
		std::string	m_result_file_name;

		//!      
		//!  .
		std::string
		make_tmp_file_name(
			//!    .
			const std::string & physic_name );
	};

//
// unclear_file_committer_t
//

/*!
	\since v.1.4.0
	\brief  %committer_t,    
		  .

	      
	 oess_1::db::storage::trace_last.   
	 ""      .

	  detect_nonactual_traces()   \a traces .
	 trace_stream_t    traces   
	 (  ).    trace-
	  .

	  initialize()  ""    .
	   :   trace- 
	      ,  c, 
	        
	trace-,         
	.     trace- 
	.    ,   
	    .     
	\a total_page_count    0  "" 
	   .

	  finalize()     
	  .

	  finalize()   (..  
	),    .

	 is_insufficient_trace_info_allowed()   true.
*/
class	OESS_1__DB__TYPE	unclear_file_committer_t
	:	public committer_t
	{
	public :
		//!  .
		unclear_file_committer_t(
			//!    .
			const std::string & physic_name );
		virtual ~unclear_file_committer_t();

	//! \name   .
	//! \{
		virtual void
		detect_nonactual_traces(
			trace_file_number_set_t & traces );

		/*!
			\note  total_page_count   ,  
				,    ""   
				.
		*/
		virtual void
		initialize(
			oess_1::uint_t total_page_count,
			oess_1::uint_t page_size,
			oess_1::uint_t block_size );

		virtual void
		finalize();

		virtual void
		store_block(
			oess_1::uint_t ordinal,
			oess_1::uint_t block_index,
			const std::string & block_value );

		virtual bool
		is_insufficient_trace_info_allowed() const;
	//! \}

	private :
		//!    .
		std::string	m_physic_name;
		//! ""  .   .
		oess_1::file::file_t	m_file;

		//!       .
		/*!
			  initialize().
		*/
		oess_1::uint_t	m_page_size;

		//!   /   store_block.
		std::vector< oess_1::char_t >	m_io_buf;
	};

//
// snapshot_file_committer_t
//

/*!
	\since v.1.4.0
	\brief  %committer_t,   
		 snapshot-.

	  detect_nonactual_traces()   trace-,
	      
	snapshot-     snapshot-.  \a traces
	   ,     snapshot
	(  ).

	     std_file_committer_t.
*/
class	OESS_1__DB__TYPE	snapshot_file_committer_t
	:	public std_file_committer_t
	{
		//!    .
		typedef std_file_committer_t	base_type_t;

	public :
		//!  .
		snapshot_file_committer_t(
			//!    .
			const std::string & physic_name );
		virtual ~snapshot_file_committer_t();

	//! \name   .
	//! \{
		virtual void
		detect_nonactual_traces(
			trace_file_number_set_t & traces );
	//! \}

		//!   trace-,    snapshot.
		/*!
			       ,  ,
			.

			   ,   
			reduce_traces_set().

			\attention      
				reduce_traces_set()        
				snapshot-.      , 
				 snapshot-  .
		*/
		trace_file_number_set_t
		outdated_traces() const;

	private :
		//!  trace-,    snapshot.	
		trace_file_number_set_t	m_outdated_traces;
	};

} /* namespace trx */

} /* namespace storage */

} /* namespace db */

} /* namespace oess */

#endif

