/*

Copyright (c) 2002-2005, Yauheni Akhotnikau
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

- The name of the author may not be used to endorse or promote products derived
from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.

*/

/*!
	\since v.1.4.0
	\brief    %trace_stream_t.
*/
#if !defined( _OESS_1__DB__STORAGE__TRX__STD_TRACE_STREAM_HPP_ )
#define _OESS_1__DB__STORAGE__TRX__STD_TRACE_STREAM_HPP_

#include <memory>

#include <oess_1/db/storage/trx/h/types.hpp>

namespace oess_1 {

namespace db {

namespace storage {

namespace trx {

namespace impl
{
	class chain_ostream_t;
} /* namespace impl */

//
// std_trace_stream_t
//

/*!
	\since v.1.4.0
	\brief    %trace_stream_t.

	      trace_all.

	 append  reverse_replay    
	    open  create.

	 is_exists  destroy     
	,   trace-stream  .
*/
class	std_trace_stream_t
	:	public trace_stream_t
	{
	public :
		std_trace_stream_t();
		virtual ~std_trace_stream_t();

	//! \name   .
	//! \{
		virtual void
		open(
			const std::string & physic_name,
			const config_t & config,
			bool is_read_only );

		virtual void
		create(
			const std::string & physic_name,
			const config_t & config );

		virtual void
		destroy(
			const std::string & physic_name );

		virtual bool
		is_exists(
			const std::string & physic_name );

		virtual void
		append(
			const trace_stream_item_t & item );

		virtual void
		reverse_replay(
			committer_t & committer,
			trace_file_number_set_t * used_files = 0,
			trace_file_number_set_t * unused_files = 0,
			std::ostream * verbose_stream = 0 );

		virtual void
		close();
	//! \}

	protected :
		//!      
		//!  trace-.
		/*!
			     , 
			     trace-stream 
			  trace-.

			\return     ,
				    
				 .
		*/
		virtual std::auto_ptr< impl::chain_ostream_t >
		create_chain_ostream();

		//!   trace-stream.
		/*!
			     
			     .
		*/
		bool
		is_open() const;

		//!        .
		/*!
			     
			     .
		*/
		const std::string &
		physic_name() const;

		//!   .
		/*!
			     
			     .
		*/
		const config_t &
		config() const;

		//!     trace-.
		/*!
			     
			     .
		*/
		trace_file_number_set_t &
		all_traces();

	private :
		//!       .
		/*!
			     , 
			   open()  create().
		*/
		std::string	m_physic_name;

		//!  .
		/*!
			     , 
			   open()  create().
		*/
		config_t	m_config;

		//!    trace-.
		/*!
			    open(), create().
			    append().
		*/
		trace_file_number_set_t	m_all_traces;

		//!  ,   .
		/*!
			 true,  trace-   
			   append()  reverse_replay.
		*/
		bool	m_is_open;

		//!  ,      .
		/*!
			   true,   
			append() .
		*/
		bool	m_is_read_only;

		//!     .
		/*!
			     append().
		*/
		std::auto_ptr< impl::chain_ostream_t >	m_chain_ostream;
	};

//
// last_trx_only_trace_stream_t
//

/*!
	\since v.1.4.0
	\brief   %trace_stream_t   
		 trace-stream   .

	    std_trace_stream_t.  
	 :
	-  chain_ostream.     
	chain_ostream,    trace-  
	;
	-  .     
	      , 
	     
	   .
*/
class	last_trx_only_trace_stream_t
	:	public std_trace_stream_t
	{
		typedef std_trace_stream_t base_type_t;
	public :
		last_trx_only_trace_stream_t();
		virtual ~last_trx_only_trace_stream_t();

	protected :
		//!      
		//!  trace-.
		/*!
			     , 
			     trace-,  
			trace- .

			\return     ,
				    
				 .
		*/
		virtual std::auto_ptr< impl::chain_ostream_t >
		create_chain_ostream();
	};

//
// pseudo_trace_stream_t
//

/*!
	\since v.1.4.0
	\brief   %trace_stream_t  , 
		    .

	,   reverse_replay,   . 
	reverse_replay  .

	 is_exists   true.
*/
class	pseudo_trace_stream_t
	:	public trace_stream_t
	{
	public :
		pseudo_trace_stream_t();
		virtual ~pseudo_trace_stream_t();

	//! \name   .
	//! \{
		virtual void
		open(
			const std::string & physic_name,
			const config_t & config,
			bool is_read_only );

		virtual void
		create(
			const std::string & physic_name,
			const config_t & config );

		virtual void
		destroy(
			const std::string & physic_name );

		virtual bool
		is_exists(
			const std::string & physic_name );

		virtual void
		append(
			const trace_stream_item_t & item );

		virtual void
		reverse_replay(
			committer_t & committer,
			trace_file_number_set_t * used_files = 0,
			trace_file_number_set_t * unused_files = 0,
			std::ostream * verbose_stream = 0 );

		virtual void
		close();
	//! \}
	};

} /* namespace trx */

} /* namespace storage */

} /* namespace db */

} /* namespace oess */

#endif

