/*

Copyright (c) 2002-2005, Yauheni Akhotnikau
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

- The name of the author may not be used to endorse or promote products derived
from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.

*/

/*!
	\since v.1.4.0
	\brief    %trace_stream_t.
*/

#include <vector>
#include <algorithm>
#include <functional>
#include <stdexcept>
#include <cstring>

#include <zlib.h>

#include <cpp_util_2/h/lexcast.hpp>
#include <cpp_util_2/lexcasts/h/util.hpp>

#include <oess_1/defs/h/ex.hpp>

#include <oess_1/io/h/mem_buf.hpp>
#include <oess_1/io/h/fixed_mem_buf.hpp>
#include <oess_1/io/h/bin_data_size.hpp>
#include <oess_1/file/h/file.hpp>

#include <oess_1/stdsn/h/ent_std.hpp>

#include <oess_1/db/storage/h/errno.hpp>

#include <oess_1/db/storage/trx/h/find_all_trace_files.hpp>
#include <oess_1/db/storage/trx/h/std_trace_stream.hpp>

#include <oess_1/debug/h/time_meter.hpp>

#define OPT_COUT(stream_ptr) !(stream_ptr) ? std::cout : *(stream_ptr)

namespace oess_1 {

namespace db {

namespace storage {

namespace trx {

namespace impl
{

/*! \name ,    
	   . */
//! \{

//! ,     .
const oess_1::uint_t	checksum_field_offset = 0;
//!   .
const oess_1::uint_t checksum_field_size = 
	oess_1::io::bin_data_size_t< oess_1::uint_t >::image_size;

//! ,     .
const oess_1::uint_t header_field_offset = checksum_field_size;
//!  .
const oess_1::uint_t header_field_size =
	oess_1::io::bin_data_size_t< oess_1::uint_t >::image_size * 7;
	
//! ,      .
const oess_1::uint_t data_field_offset =
	header_field_offset + header_field_size;
//! \}

//
// calc_checksum
//

/*!
	\since v.1.4.0
	\brief      .

	     adler32.

	\return   .
*/
oess_1::uint_t
calc_checksum(
	//! ,     .
	const oess_1::char_t * what,
	//!  .
	oess_1::uint_t size )
	{
		uLong v = adler32( 0L, Z_NULL, 0 );
		v = adler32( v,
			reinterpret_cast< const Bytef * >( what ), size );

		return static_cast< oess_1::uint_t >( v );
	}

//
// create_trace_file
//

/*!
	\since v.1.4.0
	\brief  ,  trace-.
*/
void
create_trace_file(
	//!         trace-.
	const std::string & physic_name,
	//!   trace-.
	oess_1::uint_t file_ordinal )
	{
		oess_1::file::file_t file;
		file.create( oess_1::db::storage::trx::create_trace_file_name(
				physic_name, file_ordinal ) );
	}

//
// destroy_trace_files
//

/*!
	\since v.1.4.0
	\brief   trace-.

	,   oess_1::ex_t   ,
	     \a errors.
*/
void
destroy_trace_files(
	//!       .
	const std::string & physic_name,
	//!  trace-  .
	const trace_file_number_set_t & all_traces,
	//!       .
	std::string & errors )
	{
		for( trace_file_number_set_t::const_iterator
				it = all_traces.begin(), it_end = all_traces.end();
				it != it_end; ++it )
			{
				//    , ..  
				//   -    
				// trace-.
				try
					{
						oess_1::file::file_t::remove(
								oess_1::db::storage::trx::
										create_trace_file_name(
												physic_name, *it ) );
					}
				catch( const oess_1::ex_t & x )
					{
						errors += std::string( x.what() ) + ";";
					}
			}
	}

//
// check_trace_file_count
//

/*!
	\since v.1.4.0
	\brief   trace-   .

	\par  trace_all
	     trace-.

	\par  trace_last
	    trace-   0.

	\return 0,   .  ,    
		trace-   \a trace_mode.
*/
err_code_t
check_trace_file_count(
	//!       .
	const trace_mode_t & trace_mode,
	//!  trace-   .
	const trace_file_number_set_t & traces )
	{
		if( trace_all == trace_mode )
			{
				if( 0 == traces.size() )
					{
						return oess_1_make_err(
								oess_1::db::storage::err::c_trace_file_not_found,
								"must be at least one trace file in "
										"trace_all mode" );
					}
			}
		else if( trace_last == trace_mode )
			{
				if( 1 != traces.size() || 0 != *( traces.begin() ) )
					{
						return oess_1_make_err(
								oess_1::db::storage::err::c_trace_file_not_found,
								"must be only one trace file in "
										"trace_last mode" );
					}
			}

		return err_code_t( __FILE__, __LINE__ );
	}


//
// chain_ostream_t
//

/*!
	\since v.1.4.0
	\brief  ,      
		    trace-.

	       trace- 
	 ,  chain_ostream_t   .
	  trace-     
	  trace-.  ,   
	       trace-,
	     trace-.

	    trace- ,    
	  ,  trace-  
	  .

	      
	append.

	\note     ,   
	    .    
	     .

	\par    .
\code
//     .
oess_1::uint_t	m_checksum;
//        .
oess_1::uint_t	m_chain_ordinal;
//      .
oess_1::uint_t m_chain_size;
//      .
oess_1::uint_t m_image_len;
//    .
oess_1::uint_t m_image_checksum;
oess_1::uint_t m_reserved[ 3 ];
\endcode
*/
class	chain_ostream_t
	:	protected oess_1::io::obinstream_t
	{
	public :
		//!  .
		/*!
			   trace- ,  , 
			 .
		*/
		chain_ostream_t(
			//!   ,    .
			const std::string & physic_name,
			//!    trace-.
			//!    .
			trace_file_number_set_t & trace_file_numbers,
			//!    trace-  .
			oess_1::uint_t trace_file_page_size,
			//!    trace-  .
			oess_1::uint_t trace_file_size_limit,
			//!  ,    
			//!      .
			oess_1::uint_t trace_stream_buffered_pages );
		virtual ~chain_ostream_t();

		/*!
			    .
		*/
		virtual void
		append(
			const oess_1::char_t * image,
			oess_1::uint_t image_size );

	protected :
	//! \name   .
	//! \{
		virtual oess_1::char_t *
		out_reserve(
			size_t item_count,
			size_t item_size,
			size_t & item_available );

		virtual void
		out_shift(
			size_t item_count,
			size_t item_size );
	//! \}

	protected :
		//!   ,    trace-.
		std::string	m_physic_name;
		//!    trace-.
		//!    .
		trace_file_number_set_t & m_trace_file_numbers;
		//!    trace-  .
		oess_1::uint_t	m_trace_file_page_size;
		//!    trace-  .
		const oess_1::uint_t	m_trace_file_size_limit;
		//!  trace-,    .
		oess_1::file::file_t	m_file;

	/*! \name  .

		,     
		     append().

		\{ */

		//!   .
		oess_1::uint_t m_image_len;
		//!    .
		oess_1::uint_t m_image_checksum;

		//!     .
		oess_1::uint_t	m_chain_page_ordinal;
		//!    trace-   .
		oess_1::uint_t	m_chain_page_count;

	/*! \} */

		//!   .
		/*!
			      .
			     
			 .
		*/
		std::vector< oess_1::char_t >	m_buf;
		//!      .
		/*!
			    checksum  header.
		*/
		oess_1::uint_t	m_buf_size;

		//!       .
		oess_1::io::mem_buf_t	m_buffered_pages;

		//!    .
		oess_1::uint_t
		page_size() const
			{
				return m_trace_file_page_size;
			}

		//!    .
		/*!
			       ,
			      
			   .
		*/
		void
		open_file_to_append(
			//!   trace-.
			oess_1::uint_t file_ordinal );

		//!      .
		/*!
			   m_image_len, m_image_checksum.

			  m_chain_page_ordinal.

			 m_chain_page_ordinal.
		*/
		void
		prepare_new_transaction(
			const oess_1::char_t * image,
			oess_1::uint_t image_size );

		//!     .
		/*!
			 m_buf   checksum,   
			    trace-.   
			trace-    , 
			   trace-.
		*/
		void
		push_out_current_page();

		//!    .
		/*!
			     checksum.

			  m_chain_page_ordinal  .
		*/
		void
		complete_current_page();

		//!     .
		/*!
			  trace-  ,  
			    .
		*/
		void
		flush_buffered_pages();

		//!    trace-.
		/*!
			    ,   
			   -   
			  trace-.
		*/
		virtual void
		switch_to_next_trace_file();

		//!  m_buf   .
		/*!
			 m_buf   m_buf_size , 
			     checksum.
		*/
		void
		start_new_page()
			{
				std::memset( &m_buf[ 0 ], 0, m_buf.size() );
				m_buf_size = data_field_offset;
			}

		//!   ,   
		//!       trace-stream.
		/*!
			\return    m_trace_file_numbers.
		*/
		oess_1::uint_t
		current_file_to_append() const
			{
				if( !m_trace_file_numbers.size() )
					OESS_THROW_PHYSIC(
							oess_1::db::storage::err::c_trace_file_not_found,
							"no trace files for: " << m_physic_name )

				return *( m_trace_file_numbers.rbegin() );
			}
	};

chain_ostream_t::chain_ostream_t(
	const std::string & physic_name,
	trace_file_number_set_t & trace_file_numbers,
	oess_1::uint_t trace_file_page_size,
	oess_1::uint_t trace_file_size_limit,
	oess_1::uint_t trace_stream_buffered_pages )
	:	m_physic_name( physic_name )
	,	m_trace_file_numbers( trace_file_numbers )
	,	m_trace_file_page_size( trace_file_page_size )
	,	m_trace_file_size_limit( trace_file_size_limit )
	,	m_image_len( 0 )
	,	m_image_checksum( 0 )
	,	m_chain_page_ordinal( 0 )
	,	m_chain_page_count( 0 )
	,	m_buf( trace_file_page_size, '\0' )
	,	m_buf_size( 0 )
	,	m_buffered_pages( trace_file_page_size * trace_stream_buffered_pages )
	{
		start_new_page();

		open_file_to_append( current_file_to_append() );
	}

chain_ostream_t::~chain_ostream_t()
	{}

void
chain_ostream_t::append(
	const oess_1::char_t * image,
	oess_1::uint_t image_len )
	{
		prepare_new_transaction( image, image_len );

		//    .
		write( image, image_len );

		//   .
		push_out_current_page();
		flush_buffered_pages();
	}

oess_1::char_t *
chain_ostream_t::out_reserve(
	size_t item_count,
	size_t item_size,
	size_t & item_available )
	{
		do
			{
				item_available = ( page_size() - m_buf_size ) / item_size;
				if( !item_available )
					//      .
					push_out_current_page();
			}
		while( !item_available );

		return &m_buf[ m_buf_size ];
	}

void
chain_ostream_t::out_shift(
	size_t item_count,
	size_t item_size )
	{
		m_buf_size += item_count * item_size;
	}

void
chain_ostream_t::open_file_to_append(
	oess_1::uint_t file_ordinal )
	{
		m_file.open(
				oess_1::db::storage::trx::create_trace_file_name(
						m_physic_name, file_ordinal ),
				false );

		m_file.seek( m_file.length() );

		oess_1::uint_t tail_size = m_file.length() % page_size();
		if( tail_size )
			{
				//  .
				std::vector< oess_1::char_t >	padding(
					page_size() - tail_size, '\0' );
				m_file.write( &padding[ 0 ], padding.size() );
			}
	}

void
chain_ostream_t::prepare_new_transaction(
	const oess_1::char_t * image,
	oess_1::uint_t image_size )
	{
		//        .
		m_image_len = image_size;
		m_image_checksum = calc_checksum( image, image_size );

		//          .
		m_chain_page_ordinal = 0;

		//   ,    
		//  .
		oess_1::uint_t data_size_on_page = page_size() - data_field_offset;
		m_chain_page_count = m_image_len / data_size_on_page +
				( m_image_len % data_size_on_page ? 1 : 0 );
	}

void
chain_ostream_t::push_out_current_page()
	{
		complete_current_page();

		if( m_buffered_pages.size() == m_buffered_pages.capacity() )
			flush_buffered_pages();

		m_buffered_pages.write( &m_buf[ 0 ], page_size() );

		start_new_page();
	}

void
chain_ostream_t::complete_current_page()
	{
		oess_1::io::ofixed_mem_buf_t ostream( &m_buf[ 0 ], page_size() );

		oess_1::uint_t checksum_value = 0;

		//optimization:      
		//  write.
		oess_1::uint_t all_header[] =
			{
				checksum_value, m_chain_page_ordinal, m_chain_page_count,
				m_image_len, m_image_checksum
			};
		ostream.write( all_header,
				sizeof( all_header ) / sizeof( all_header[ 0 ] ) );

		ostream.set_pos( header_field_offset );
		checksum_value = calc_checksum(
			ostream.current_ptr(), page_size() - header_field_offset );

		ostream.set_pos( checksum_field_offset );
		ostream << checksum_value;

		++m_chain_page_ordinal;
	}

void
chain_ostream_t::flush_buffered_pages()
	{
		oess_1::uint_t remaining = m_buffered_pages.size();
		const oess_1::char_t * page = m_buffered_pages.ptr();

		while( remaining )
			{
				if( m_file.length() >= m_trace_file_size_limit )
					{
						//     trace-.
						switch_to_next_trace_file();
					}
				else
					{
						oess_1::uint_t portion_size = m_trace_file_size_limit -
								m_file.length();
						if( portion_size > remaining )
							portion_size = remaining;

						m_file.write( page, portion_size );

						page += portion_size;
						remaining -= portion_size;
					}
			}

		m_buffered_pages.change_size( 0 );
	}

void
chain_ostream_t::switch_to_next_trace_file()
	{
		oess_1::uint_t file = current_file_to_append() + 1;
		create_trace_file( m_physic_name, file );
      
		//       .
		m_trace_file_numbers.insert( file );
      
		m_file.close();
		open_file_to_append( file );
	}

//
// last_trx_only_chain_ostream_t
//

/*!
	\since v.1.4.0
	\brief  ,      
		 .

	      
	  oess_1::db::storage::trace_last.

	 trace-stream    ,     
	chain_ostream_t.    ,    
	:

	-  trace-     (..  
	 );
	-     trace-.  trace- 
	, ..   trace-    .
*/
class	last_trx_only_chain_ostream_t
	:	public chain_ostream_t
	{
		typedef chain_ostream_t base_type_t;

	public :
		//!  .
		/*!
			 trace_file_numbers     
			     .   
			 .
		*/
		last_trx_only_chain_ostream_t(
			//!   ,    .
			const std::string & physic_name,
			//!    trace-.
			//!    .
			trace_file_number_set_t & trace_file_numbers,
			//!    trace-  .
			oess_1::uint_t trace_file_page_size,
			//!  ,    
			//!      .
			oess_1::uint_t trace_stream_buffered_pages );
		virtual ~last_trx_only_chain_ostream_t();

		/*!
			    .
		*/
		virtual void
		append(
			const oess_1::char_t * image,
			oess_1::uint_t image_size );

	protected :
		//!      trace-.
		/*!
			 , ..      
			 trace-.
		*/
		virtual void
		switch_to_next_trace_file();

		//!    trace-  .
		void
		truncate_trace_file();
	};

last_trx_only_chain_ostream_t::last_trx_only_chain_ostream_t(
	const std::string & physic_name,
	trace_file_number_set_t & trace_file_numbers,
	oess_1::uint_t trace_file_page_size,
	oess_1::uint_t trace_stream_buffered_pages )
	:	base_type_t(
			physic_name, trace_file_numbers,
			trace_file_page_size,
			static_cast< oess_1::uint_t >( -1 ),
			trace_stream_buffered_pages )
	{
		err_code_t error = check_trace_file_count( trace_last,
				trace_file_numbers );

		if( error )
			OESS_THROW_PHYSIC( error.m_code, error.m_desc )
	}

last_trx_only_chain_ostream_t::~last_trx_only_chain_ostream_t()
	{}

void
last_trx_only_chain_ostream_t::append(
	const oess_1::char_t * image,
	oess_1::uint_t image_len )
	{
		truncate_trace_file();

		//     .
		base_type_t::append( image, image_len );
	}

void
last_trx_only_chain_ostream_t::switch_to_next_trace_file()
	{
		OESS_THROW_PHYSIC(
				oess_1::db::storage::err::c_trace_stream_full,
				"all transaction must fit into one trace file in "
				"trace_last mode" )
	}

void
last_trx_only_chain_ostream_t::truncate_trace_file()
	{
		m_file.seek( 0 );
		m_file.chsize( 0 );
	}

//
// reverse_loader_t
//

/*!
	\since v.1.4.0
	\brief      trace-stream  
		.
*/
class reverse_loader_t
	:	private cpp_util_2::nocopy_t
	{
	public :
		//!  .
		reverse_loader_t(
			//!       .
			const std::string & physic_name,
			//!   trace-  .
			oess_1::uint_t page_size,
			//!    trace-.
			const trace_file_number_set_t & all_traces,
			//!      
			//!   .
			std::ostream * verbose_stream = 0 );
		virtual ~reverse_loader_t();

		//!    trace-   
		//!  trace-.
		/*!
			 ,   next()      
			   ,  
			,  trace-  
			(   snapshot-).   
			  trace- ,   
			 . ,   
			  ,  ,  
			  m_all_traces.
		*/
		void
		reduce_traces_set(
			//!   trace-,  
			//!   .   ,  ,
			//!     trace-.
			const trace_file_number_set_t & nonactual_traces );

		//!   .
		/*!
			     
			   trace-.  	
			 --   ..,    
			 trace-stream.	

			\attention   \a image  
			    next()  
			  reverse_loader_t.
		*/
		virtual bool
		next(
			//!     .
			const oess_1::char_t * & image,
			//!   .
			oess_1::uint_t & image_size,
			//!   trace-   
			//!   . ,  
			//! trace-,     .
			trace_file_number_set_t & used_traces );

	private :
		//!       .
		std::string	m_physic_name;
		//!   trace-.
		oess_1::uint_t	m_page_size;

		//!      
		//!   .
		std::ostream * m_verbose_stream;

		//!   ,   
		//!  trace-stream.
		/*!
			       	
			\a all_traces . ,    
			trace-stream      
			.  ,    
			.
		*/
		trace_file_number_set_t	m_all_traces;

		//!  ,    .
		bool	m_is_current_file_open;
		//!   trace-.
		/*!
			   , 
			m_is_current_file_open == true.
		*/
		trace_file_number_t	m_current_file_number;
		//!     trace-.
		/*!
			   , 
			m_is_current_file_open == true.
		*/
		oess_1::uint_t	m_current_file_page;
		/*!
			  .

			   , 
			m_is_current_file_open == true.
		*/
		std::vector< oess_1::char_t >	m_current_file_content;

	//! \name ,     .
	//! \{
		//! ,     .
		/*!
			      
			     (, 
			     
			  ).
		*/
		std::vector< oess_1::char_t >	m_image;

		//!      .
		/*!
			     
			,    1  
			  .

			  m_current_ordinal ,  m_current_ordinal
			  ,   m_current_page.
			  m_chain_page_ordinal  
			 reverse_loader_t.
		*/
		oess_1::uint_t	m_chain_page_ordinal;

		//!   .
		/*!
			     ,
			      .
		*/
		oess_1::uint_t	m_image_len;

		//!    .
		/*!
			     ,
			      .
		*/
		oess_1::uint_t	m_image_checksum;

		/*!
			  trace-,   
			    .

			    .
		*/
		trace_file_number_set_t	m_used_traces;
	//! \}

	//! \name ,      .
	//! \{
		//!     .
		oess_1::uint_t	m_current_ordinal;

		//!     .
		oess_1::uint_t m_current_count;

		//!   .
		oess_1::uint_t	m_current_image_len;

		//!    .
		oess_1::uint_t	m_current_image_checksum;
	//! \}

		//!    trace-.
		/*!
			  m_all_traces    
			   .   
			  trace-  m_all_traces
			.

			    ( 
			),      m_all_traces.

			      
			m_current_file_number.

			  m_all_traces  , 
			  end_of_trace_ex_t.

			   ,    
			  m_current_file_content.
		*/
		void
		try_open_next_file();

		//!   .
		/*!
			  ,    ,  image_type,
			  m_image.

			\return true,   .
				false,   trace-    
				 .
		*/
		void
		find_trx();

		//!      .
		void
		reinit_search();

		//!    .
		/*!
			   ,   
			,    .

			      m_image.

			    ,  
			  .    
			   .
		*/
		void
		find_first_trx_page();

		//!   .
		/*!
			     
			find_first_trx_page().

			       
			,      .
		*/
		void
		pop_up_trx();

		//!        .
		/*!
			\throw trace_file::broken_page_ex_t,    
				       .
		*/
		void
		pop_up_and_check_page();

		//!      trace-.
		/*!
			  m_trx_current   
			  .    , 
			  trace-,    
			 trace-.

			   trace-  
			end_of_trace_ex_t.
		*/
		void
		go_to_next_page();

		//!         trace.
		/*!
			  go_to_next_page    trace-,
			  end_of_trace_ex_t  .
		*/
		void
		go_to_next_page_no_throw_at_end();

		//!  ,      %m_image.
		bool
		is_all_trx_loaded() const;

		//!    %m_image.	
		/*!
			    .

			\throw broken_trx_t,     m_image
				 .
		*/
		void
		check_trx_integrity() const;

		//!  ,    
		//!  .
		bool
		is_current_trx_page() const;

		//!       %m_image.
		/*!
			     ,  
			    m_used_traces.
		*/
		void
		store_page_data();

		//!     trace-.
		oess_1::uint_t
		page_size() const;
	};

//! \name    reverse_loader_t.
//! \{

//
//     reverse_loader_t  ,
//     std::logic_error ( ), 
//   oess_1::ex_t.
//

//!    trace-.
class	end_of_trace_ex_t
	:	public std::logic_error
	{
	typedef std::logic_error base_type_t;
	public :
		end_of_trace_ex_t()
			:	base_type_t( "end of all trace stream" )
			{}	
	};

//!  trace-   .
class	broken_page_ex_t
	:	public std::logic_error
	{
	typedef std::logic_error base_type_t;
	public :
		broken_page_ex_t()
			:	base_type_t( "broken page extracted" )
			{}	
	};

//!  trace-   .
class	broken_trx_ex_t
	:	public std::logic_error
	{
	typedef std::logic_error base_type_t;
	public :
		broken_trx_ex_t()
			:	base_type_t( "broken trx extracted" )
			{}	
	};

//! \}

//
// reverse_loader_t
//

reverse_loader_t::reverse_loader_t(
	const std::string & physic_name,
	oess_1::uint_t page_size,
	const trace_file_number_set_t & all_traces,
	std::ostream * verbose_stream )
	:	m_physic_name( physic_name )
	,	m_page_size( page_size )
	,	m_all_traces( all_traces )
	,	m_verbose_stream( verbose_stream )
	,	m_is_current_file_open( false )
	{
	}

reverse_loader_t::~reverse_loader_t()
	{}

void
reverse_loader_t::reduce_traces_set(
	const trace_file_number_set_t & nonactual_traces )
	{
		if( !nonactual_traces.empty() )
			{
				trace_file_number_set_t updated;
				std::set_difference(
						m_all_traces.begin(), m_all_traces.end(),
						nonactual_traces.begin(), nonactual_traces.end(),
						std::inserter( updated, updated.begin() ) );
				updated.swap( m_all_traces );
			}
	}

bool
reverse_loader_t::next(
	const oess_1::char_t * & image,
	oess_1::uint_t & image_size,
	trace_file_number_set_t & used_traces )
	{
		try
			{
				if( !m_is_current_file_open )
					try_open_next_file();

				find_trx();

				image_size = m_image_len;
				image = &m_image[ 0 ];
				used_traces = m_used_traces;

				return true;
			}
		catch( const end_of_trace_ex_t & )
			{}

		return false;
	}

void
reverse_loader_t::try_open_next_file()
	{
		while( !m_is_current_file_open && m_all_traces.size() )
			{
				trace_file_number_t number = *( m_all_traces.rbegin() );

				std::string file_name = oess_1::db::storage::trx::
						create_trace_file_name( m_physic_name, number );
				if( oess_1::file::file_t::is_file_exists( file_name ) )
					{
						oess_1::file::file_t file;
						file.open( file_name, true );
						//     :
						//     ...
						m_current_file_content.resize( file.length(), 0 );
						file.read( &m_current_file_content[ 0 ],
								m_current_file_content.size() );

						// ...     .
						m_current_file_page =
								m_current_file_content.size() / page_size();
						if( m_current_file_page &&
								!( m_current_file_content.size() % page_size() ) )
							//     ,   
							//        
							//   .    
							//      . 
							//      1.
							--m_current_file_page;

						m_current_file_number = number;
						m_is_current_file_open = true;

						OPT_COUT( m_verbose_stream )
								<< file_name << ": trace-file open; "
								<< "current_file_page: " << m_current_file_page
								<< std::endl;
					}
				else
					OPT_COUT( m_verbose_stream )
							<< file_name << ": not exists; ignored"
							<< std::endl;

				m_all_traces.erase( number );
			}

		if( !m_is_current_file_open )
			{
				//    ,    .
				throw end_of_trace_ex_t();
			}
	}

void
reverse_loader_t::find_trx()
	{
		//     .
		//     .
		//    .
		while( true )
			{
				//    .
				reinit_search();

				try
					{
						find_first_trx_page();
						pop_up_trx();

						return;
					}
				//       .
				//        
				//   ,    .
				catch( const broken_page_ex_t & )
					{
						//    , 
						//    .
						go_to_next_page();
					}
				catch( const broken_trx_ex_t & )
					{
					}
			}
	}

void
reverse_loader_t::reinit_search()
	{
		m_used_traces.clear();
	}

void
reverse_loader_t::find_first_trx_page()
	{
		//   ,    
		// ,       
		//     -- 
		//       .
		while( true )
			{
				pop_up_and_check_page();
				if( m_current_ordinal + 1 == m_current_count )
					{
						m_image_len = m_current_image_len;
						m_image_checksum = m_current_image_checksum;

						//     .
						m_chain_page_ordinal = m_current_ordinal;

						//        .
						if( m_image.size() < m_image_len )
							m_image.resize( m_image_len, 0 );

						//      ,
						// .. pop_up_trx    .
						store_page_data();

						OPT_COUT( m_verbose_stream )
								<< m_physic_name << ":"
								<< m_current_file_number << ":"
								<< m_current_file_page << ": "
								<< "start of transaction found: "
								<< "page count: " << m_current_count
								<< ", image length: " << m_image_len
								<< ", image checksum: 0x"
								<< std::hex << m_image_checksum
								<< std::dec << std::endl;

						return;
					}
				else
					{
						OPT_COUT( m_verbose_stream )
								<< m_physic_name << ":"
								<< m_current_file_number << ":"
								<< m_current_file_page << ": "
								<< "part of broken transaction; "
								<< "page ordinal: " << m_current_ordinal
								<< "page count: " << m_current_count
								<< ", image length: " << m_image_len
								<< ", image checksum: 0x"
								<< std::hex << m_image_checksum
								<< std::dec << std::endl;

						go_to_next_page();
					}
			}
	}

void
reverse_loader_t::pop_up_trx()
	{
		//   ,    
		//   m_image .
		while( true )
			{
				if( is_all_trx_loaded() )
					{
						//       
						// .      .
						go_to_next_page_no_throw_at_end();

						check_trx_integrity();

						// ,   .
						return;
					}
				else
					{
						//     .
						go_to_next_page();

						// ,       .
						if( m_chain_page_ordinal )
							--m_chain_page_ordinal;
						else
							//   , ..  
							//       
							//    .
							throw broken_trx_ex_t();

						pop_up_and_check_page();
						if( is_current_trx_page() )
							store_page_data();
						else
							//   , ..  
							// ,    .
							throw broken_trx_ex_t();
					}
			}
	}

void
reverse_loader_t::pop_up_and_check_page()
	{
		// -  go_to_next_page_no_throw_at_end() 
		// ,     .
		if( !m_is_current_file_open )
			throw end_of_trace_ex_t();

		oess_1::uint_t page_offset = m_current_file_page * page_size();

		if( page_offset + page_size() <= m_current_file_content.size() )
			{
				//    .
				//     .
				oess_1::io::ifixed_mem_buf_t istream(
						&m_current_file_content[ page_offset ], page_size() );
      
				//    .
				oess_1::uint_t page_checksum;
				istream >> page_checksum;
      
				oess_1::uint_t calculated_checksum = calc_checksum(
						istream.current_ptr(),
						page_size() - checksum_field_size );
				if( page_checksum != calculated_checksum )
					{
						OPT_COUT( m_verbose_stream ) << m_physic_name
								<< ":" << m_current_file_number
								<< ":" << m_current_file_page
								<< ": page checksum mismatch; "
								<< "header checksum: 0x"
								<< std::hex << page_checksum << std::dec
								<< ", real checksum: 0x"
								<< std::hex << calculated_checksum << std::dec
								<< std::endl;
						throw broken_page_ex_t();
					}
      
				//  ,      .
				istream >> m_current_ordinal
					>> m_current_count
					>> m_current_image_len
					>> m_current_image_checksum;
			}
		else
			{
				OPT_COUT( m_verbose_stream ) << m_physic_name
						<< ":" << m_current_file_number
						<< ":" << m_current_file_page
						<< ": not all page in file; "
						<< "page offset: " << page_offset
						<< ", file length: " << m_current_file_content.size()
						<< std::endl;

				//   , ..    .
				throw broken_page_ex_t();
			}
	}

void
reverse_loader_t::go_to_next_page()
	{
		// -  go_to_next_page_no_throw_at_end() 
		// ,     .
		if( !m_is_current_file_open )
			throw end_of_trace_ex_t();

		if( !m_current_file_page )
			{
				//     .
				m_is_current_file_open = false;

				try_open_next_file();
			}
		else
			//     .
			--m_current_file_page;
	}

void
reverse_loader_t::go_to_next_page_no_throw_at_end()
	{
		try
			{
				go_to_next_page();
			}
		catch( const end_of_trace_ex_t & )
			{}
	}

bool
reverse_loader_t::is_all_trx_loaded() const
	{
		return ( 0 == m_current_ordinal );
	}

void
reverse_loader_t::check_trx_integrity() const
	{
		oess_1::uint_t image_calculated_checksum = calc_checksum(
				&m_image[ 0 ], m_image_len );
		if( m_image_checksum != image_calculated_checksum )
			{
				OPT_COUT( m_verbose_stream ) << m_physic_name
						<< ":" << m_current_file_number
						<< ":" << m_current_file_page
						<< ": image checksum mismatch; "
						<< "header checksum: 0x"
						<< std::hex << m_image_checksum << std::dec
						<< ", real checksum: 0x"
						<< std::hex << image_calculated_checksum << std::dec
						<< std::endl;

				throw broken_trx_ex_t();
			}
	}

bool
reverse_loader_t::is_current_trx_page() const
	{
		if( m_chain_page_ordinal == m_current_ordinal )
			if( m_image_len == m_current_image_len &&
					m_image_checksum == m_current_image_checksum )
				return true;

		OPT_COUT( m_verbose_stream ) << m_physic_name
				<< ":" << m_current_file_number
				<< ":" << m_current_file_page
				<< ": is not a part of current transaction; "
				<< " expected ordinal: " << m_chain_page_ordinal
				<< ", page ordinal: " << m_current_ordinal
				<< ", page count: " << m_current_count
				<< ", expected image length: " << m_image_len
				<< ", image length: " << m_current_image_len
				<< ", expected image checksum: 0x"
				<< std::hex << m_image_checksum << std::dec
				<< ", image checksum: 0x"
				<< std::hex << m_current_image_checksum << std::dec
				<< std::endl;
		
		return false;
	}

void
reverse_loader_t::store_page_data()
	{
		oess_1::uint_t page_offset = m_current_file_page * page_size();

		oess_1::uint_t data_bytes_on_page = ( page_size() -
				data_field_offset );
		oess_1::uint_t pos = m_current_ordinal * data_bytes_on_page;

		if( pos + data_bytes_on_page > m_image_len )
			data_bytes_on_page = m_image_len - pos;

		std::copy(
				&m_current_file_content[
						page_offset + data_field_offset ],
				&m_current_file_content[
						page_offset + data_field_offset +
								data_bytes_on_page ],
				&m_image[ 0 ] + pos );

		//     -   ,
		//        .
		m_used_traces.insert( m_current_file_number );
	}

oess_1::uint_t
reverse_loader_t::page_size() const
	{
		return m_page_size;
	}

//FIX ME:       .  
// serialize_item, deserialize_item    
/// .     .
//           
//- .      
//std_trace_stream_t.  ,    
//        /.

//
// initialize_compression
//

/*!
	,      , 
	    .
*/
void
initialize_compression( const config_t & config )
	{
	}

//
// use_zlib_compression
//
 
void
use_zlib_compression(
	//!  .
	const oess_1::char_t * what,
	//!  .
	oess_1::uint_t uncompressed_size,
	//!  .
	std::vector< char_t > & compressed_data,
	//!   .
	oess_1::uint_t & compressed_size )
	{
		compressed_data.reserve( compressBound( uncompressed_size ) );
		uLong dest_length = compressed_data.capacity();
		int z_result;
		z_result = compress2(
				reinterpret_cast< Bytef * >( &compressed_data.front() ),
				&dest_length,
				reinterpret_cast< const Bytef * >( what ),
				uncompressed_size,
				Z_BEST_SPEED );
		if( Z_OK != z_result )
			OESS_THROW_PHYSIC(
					oess_1::db::storage::err::c_trx_zlib_failure,
					"try to compress: " << uncompressed_size
					<< " bytes; compress() return code: " << z_result )

		compressed_size = static_cast< oess_1::uint_t >( dest_length );
	}

//
// serialize_item
//

/*!
	\brief   %trace_stream_image_t 
		     
		 trace-stream.

	    
	all_zlib_compressed.    
	  ,    .
	     .
*/
void
serialize_item(
	//!  .
	const trace_stream_item_t & item,
	//!  ,   .
	const trace_compression_mode_t & required_compression_mode,
	//!   .
	oess_1::io::mem_buf_t & result_receiver )
	{
		//    item  ,  
		//  .
		oess_1::io::mem_buf_t item_image(
				static_cast< oess_1::uint_t >(
						1.2 * item.summary_length() ) );
		oess_1::stdsn::oent_std_t item_ent( item_image );
		item_ent << item;

		image_header_t::compression_mode_t compression_mode =
				trace_no_compression;
		oess_1::uint_t uncompressed_size = item_image.size();
		oess_1::uint_t compressed_size = uncompressed_size;
		const oess_1::char_t * image_to_store = item_image.ptr();

		std::vector< char_t > compressed_data;

		//   ,    .
		if( trace_no_compression != required_compression_mode )
			{
				if( trace_zlib_compression == required_compression_mode )
					use_zlib_compression( item_image.ptr(), uncompressed_size,
							compressed_data, compressed_size );
				else
					OESS_THROW_PHYSIC(
							oess_1::db::storage::err::c_trx_unknown_compression_mode,
							static_cast< oess_1::uint_t >( required_compression_mode )
							<< ": compression mode unknown or not implemented yet" )

				//        .
				//   ,    .
				if( compressed_size < uncompressed_size )
					{
						//      .
						compression_mode = required_compression_mode;
						image_to_store = reinterpret_cast< const oess_1::char_t * >(
								&compressed_data[ 0 ] );
					}
				else
					//   , ..  
					//       .
					compressed_size = uncompressed_size;
			}

		image_header_t header( compression_mode, uncompressed_size,
				compressed_size );

		result_receiver.change_capacity( 32 /*  
				  header */
				+ compressed_size );
		oess_1::stdsn::oent_std_t result_ent( result_receiver );
		result_ent << header;
		result_receiver.write( image_to_store, compressed_size );
	}

//
// unpack_all_zlib_compressed
//

/*!
	\brief   %trace_stream_image_t 
		   %all_zlib_compressed .
*/
void
unpack_all_zlib_compressed(
	//!   .
	const oess_1::char_t * image,
	//!   .
	oess_1::uint_t compressed_size,
	//!   .
	oess_1::uint_t uncompressed_size,
	//!   .
	trace_stream_item_t & item )
	{
		std::vector< Bytef > unpacked_data( uncompressed_size );
		uLong dest_length = uncompressed_size;
		int z_result = uncompress( &unpacked_data.front(), &dest_length,
				reinterpret_cast< const Bytef * >( image ),
				compressed_size );
		if( Z_OK != z_result )
			OESS_THROW_PHYSIC(
					oess_1::db::storage::err::c_trx_zlib_failure,
					"try to uncompress: " << compressed_size
					<< " bytes, awaiting uncompressed size: "
					<< uncompressed_size
					<< "; uncompress() return code: " << z_result )
		if( uncompressed_size != dest_length )
			OESS_THROW_PHYSIC(
					oess_1::db::storage::err::c_unexpected,
					"data length mismatch after decompression; "
					"awaiting uncompressed_size: "
					<< uncompressed_size
					<< ", zlib uncompressed length: "
					<< dest_length )

		//    .
		oess_1::io::ifixed_mem_buf_t stream(
				reinterpret_cast< const oess_1::char_t * >(
						&unpacked_data.front() ),
				uncompressed_size );
		oess_1::stdsn::ient_std_t ient( stream );
		ient >> item;
	}

//
// deserialize_item
//

/*!
	\breif   %trace_stream_image_t 
		  trace-stream  .
*/
void
deserialize_item(
	//!  .
	const oess_1::char_t * image,
	//!   .
	oess_1::uint_t image_len,
	//!   .
	trace_stream_item_t & item )
	{
		oess_1::io::ifixed_mem_buf_t stream( image, image_len );
		oess_1::stdsn::ient_std_t ient( stream );

		//   ,  ,  
		//  .
		image_header_t image_header;
		ient >> image_header;

		if( trace_no_compression == image_header.compression_mode() )
			//    ,  
			//       .
			ient >> item;
		else if( trace_zlib_compression == image_header.compression_mode() )
			unpack_all_zlib_compressed(
					stream.current_ptr(),
					image_header.compressed_size(),
					image_header.uncompressed_size(),
					item );
		else
			OESS_THROW_PHYSIC(
					oess_1::db::storage::err::c_trx_unknown_compression_mode,
					"compression mode: " << static_cast< oess_1::uint_t >(
							image_header.compression_mode() ) )
	}

//
// replay_completion_map_t
//

/*!
	\brief      
		  .

	\attention      , 	
		       8 ,  
		  .  
	\todo  ,  ,   
		   32.     
		,      
		m_masks.

	\note     ,
		    is_*  .
*/
class	replay_completion_map_t
	:	public cpp_util_2::nocopy_t
	{
	public :
		//!  .
		/*!
			\throw oess_1::physic_ex_t,  page_size / block_size > 8.
		*/
		replay_completion_map_t(
			//!    .
			oess_1::uint_t page_count,
			//!     .
			oess_1::uint_t page_size,
			//!     .
			oess_1::uint_t block_size )
			:	m_all_block_count( page_count * ( page_size / block_size ) )
			,	m_completed_blocks( 0 )
			,	m_page_block_count( page_size / block_size )
			{
				//  ,   
				//  ,   
				//   .
				switch( m_page_block_count )
					{
						case 1 : m_completed_page_mask = 0x1; break;
						case 2 : m_completed_page_mask = 0x3; break;
						case 3 : m_completed_page_mask = 0x7; break;
						case 4 : m_completed_page_mask = 0xf; break;
						case 5 : m_completed_page_mask = 0x1f; break;
						case 6 : m_completed_page_mask = 0x3f; break;
						case 7 : m_completed_page_mask = 0x7f; break;
						case 8 : m_completed_page_mask = 0xff; break;
						default :
							OESS_THROW_PHYSIC(
									oess_1::db::storage::err::
											c_invalid_page_block_count,
									m_page_block_count
									<< ": page_block_count must >= 1 and <= 8" )
					}

				m_masks.resize( page_count, 0 );
			}

		//!    .
		/*!
			\return false,     .
		*/
		bool
		is_need_continue() const
			{
				return ( m_all_block_count != m_completed_blocks );
			}

		//!      .
		/*!
			\return false,     
				.
		*/
		bool
		is_page_need_completion(
			//!     .
			oess_1::uint_t page_ordinal ) const
			{
				return ( m_completed_page_mask !=
						( m_completed_page_mask & m_masks[ page_ordinal ] ) );
			}

		//!       
		//!  .
		/*!
			\return false,     .
		*/
		bool
		is_page_fragment_need_value(
			//!     .
			oess_1::uint_t page_ordinal,
			//!    .
			oess_1::uint_t block_index ) const
			{
				oess_1::uchar_t bit = ( 0x1 << block_index );
				return !( bit & m_masks[ page_ordinal ] );
			}

		//!   ,   
		//!     .
		void
		page_fragment_defined(
			//!     .
			oess_1::uint_t page_ordinal,
			//!    .
			oess_1::uint_t block_index )
			{
				oess_1::uchar_t bit = ( 0x1 << block_index );
				m_masks[ page_ordinal ] |= bit;

				//!     .
				++m_completed_blocks;
			}

		//!      .
		void
		dump( std::ostream & to );

	private :
		//!      .
		oess_1::uint_t	m_all_block_count;
		//!    1 .
		oess_1::uint_t	m_completed_blocks;
		//!     .
		oess_1::uint_t	m_page_block_count;
		//! ,     .
		/*!       . */
		oess_1::uchar_t m_completed_page_mask;

		//!      .
		std::vector< oess_1::uchar_t >	m_masks;
	};

void
replay_completion_map_t::dump( std::ostream & to )
	{
		to << "replay_completion_map status: "
			<< "\tall_block_count: " << m_all_block_count << "\n"
			<< "\tcompleted_blocks: " << m_completed_blocks << "\n"
			<< "\tpage_block_count: " << m_page_block_count << "\n"
			<< "\tcompleted_page_mask: 0x" << std::hex
				<< static_cast< oess_1::ushort_t >( m_completed_page_mask )
				<< std::dec << "\n";

		for( oess_1::uint_t i = 0,
				i_max = m_all_block_count / m_page_block_count;
				i != i_max; )
			{
				char sz[ 32 ];
				std::sprintf( sz, "%08X", i );
				to << sz << ": ";

				for( oess_1::uint_t j = 0; j != 16 && i != i_max; ++j, ++i )
					{
						char hex_mask[ 16 ];
						std::sprintf( hex_mask, "%02x ",
								static_cast< unsigned int >( m_masks[ i ] ) );
						to << hex_mask;
					}

				to << "\n";
			}

		to << std::flush;
	}

//
// replay_context_t
//

/*!
	\brief ,     .
*/
class	replay_context_t
	:	private cpp_util_2::nocopy_t
	{
	private :
		//! ,    
		//!  .
		committer_t & 	m_committer;
		//!      .
		oess_1::uint_t	m_page_size;
		//!       .
		oess_1::uint_t m_block_size;
		//!      
		//!  .
		std::ostream * m_verbose_stream;

		//!   .
		replay_completion_map_t	m_completion;

	public :
		//!  .
		replay_context_t(
			//! ,    
			//!  .
			committer_t & committer,
			//!      .
			oess_1::uint_t total_page_count,
			//!     .
			oess_1::uint_t page_size,
			//!     .
			oess_1::uint_t block_size,
			//!      
			//!  .
			std::ostream * verbose_stream )
			:	m_committer( committer )
			,	m_page_size( page_size )
			,	m_block_size( block_size )
			,	m_verbose_stream( verbose_stream )
			,	m_completion( total_page_count, page_size, block_size )
			{}

		//!       .
		oess_1::uint_t
		page_size() const
			{
				return m_page_size;
			}

		//!         .
		oess_1::uint_t
		block_size() const
			{
				return m_block_size;
			}

		//!    .
		/*!
			\return false,     .
		*/
		bool
		is_need_continue() const
			{
				return m_completion.is_need_continue();
			}

		//!      .
		/*!
			\return false,     
				.
		*/
		bool
		is_page_need_completion(
			//!     .
			oess_1::uint_t page_ordinal ) const
			{
				return m_completion.is_page_need_completion( page_ordinal );
			}

		//!       
		//!  .
		/*!
			\return false,     .
		*/
		bool
		is_page_fragment_need_value(
			//!     .
			oess_1::uint_t page_ordinal,
			//!    .
			oess_1::uint_t block_index ) const
			{
				return m_completion.is_page_fragment_need_value(
						page_ordinal, block_index );
			}

		//!   ,   
		//!     .
		void
		page_fragment_defined(
			//!     .
			oess_1::uint_t page_ordinal,
			//!    .
			oess_1::uint_t block_index )
			{
				m_completion.page_fragment_defined(
						page_ordinal, block_index );
			}

		//!    -.
		committer_t &
		committer()
			{
				return m_committer;
			}

		//!      .
		void
		dump()
			{
				if( m_verbose_stream )
					m_completion.dump( *m_verbose_stream );
			}
	};

//
// trx_page_delta_processor_t
//

/*!
	  std::for_each.

	    
	  .
*/
class	trx_page_delta_processor_t
	:	public std::unary_function< const delta_info_shptr_t &, void >
	{
		//!     .
		replay_context_t & m_ctx;
		//!     true,   
		//!     .
		bool & m_is_applied;
		//!    .
		oess_1::uint_t	m_page_ordinal;

	public :
		trx_page_delta_processor_t(
			replay_context_t & ctx,
			bool & is_applied,
			oess_1::uint_t page_ordinal )
			:	m_ctx( ctx )
			,	m_is_applied( is_applied )
			,	m_page_ordinal( page_ordinal )
			{}

		result_type
		operator()( argument_type a )
			{
				const std::string & delta = a->value();
				oess_1::uint_t block_count = delta.size() /
						m_ctx.block_size();

				for( oess_1::uint_t block_index = a->index(), i = 0;
						i != block_count;
						++block_index, ++i )
					if( m_ctx.is_page_fragment_need_value(
							m_page_ordinal, block_index ) )
						{
							m_ctx.committer().store_block(
									m_page_ordinal,
									block_index,
									delta.substr( i * m_ctx.block_size(),
											m_ctx.block_size() ) );
							m_ctx.page_fragment_defined( m_page_ordinal,
									block_index );
							m_is_applied = true;
						}
			}
	};

//
// trx_page_processor_t
//

/*!
	  std::for_each.

	    .
*/
class	trx_page_processor_t
	:	public std::unary_function< const page_info_t &, void >
	{
		//!     .
		replay_context_t & m_ctx;
		//!     true,   
		//!     .
		bool & m_is_applied;

	public :
		trx_page_processor_t(
			replay_context_t & ctx,
			bool & is_applied )
			:	m_ctx( ctx )
			,	m_is_applied( is_applied )
			{}

		result_type
		operator()( argument_type a )
			{
				if( m_ctx.is_page_need_completion( a.ordinal() ) )
					{
						std::for_each( a.deltas().begin(), a.deltas().end(),
								trx_page_delta_processor_t(
										m_ctx, m_is_applied, a.ordinal() ) );
					}
			}
	};

//
// do_recover_db_with_damaged_last_trx
//

/*!
	\brief   ,    
		trace_last    trace-    .

	     commiter-    :
	- initialize(),    
	- finalize()  .

	    , ..    
	 "" .
*/
static void
do_recover_db_with_damaged_last_trx(
	committer_t & committer,
	const config_t & config )
	{
		committer.initialize(
				0,
				config.page_size(),
				config.trx_delta_block_size() );
		committer.finalize();
	}

//
// process_trx
//

/*!
	\brief   .

	\return true,      -
		 .
*/
static bool
process_trx(
	//!  ,   .
	const trace_stream_item_t & trx,
	//! ,    .
	replay_context_t & ctx )
	{
		bool is_applied = false;

		std::for_each( trx.pages().begin(), trx.pages().end(),
				trx_page_processor_t( ctx, is_applied ) );

		return is_applied;
	}

//
// calc_used_unused_traces
//

/*!
	\brief      
		trace-.
*/
static void
calc_used_unused_traces(
	//!    trace-  trace-stream.
	const trace_file_number_set_t & all_traces,
	//!  trace-,   
	//!   .
	const trace_file_number_set_t & all_used_traces,
	//!     trace-,
	//!       .
	trace_file_number_set_t * used_traces,
	//! e    trace-,
	//!     ,   
	//! .
	trace_file_number_set_t * unused_traces )
	{
		if( used_traces )
			*used_traces = all_used_traces;

		if( unused_traces )
			{
				unused_traces->clear();
				std::set_difference(
						all_traces.begin(), all_traces.end(),
						all_used_traces.begin(), all_used_traces.end(),
						std::inserter( *unused_traces,
								unused_traces->begin() ) );
			}
	}

//
// do_normal_recover
//

/*!
	\brief    ,   
		 trace_all.
*/
static void
do_normal_recover(
	//!       .
	const std::string & physic_name,
	//!  .
	const config_t & config,
	//!    trace-  trace-stream.
	const trace_file_number_set_t & all_traces,
	//! ,      .
	committer_t & committer,
	//!     trace-,
	//!       .
	trace_file_number_set_t * used_traces,
	//! e    trace-,
	//!     ,   
	//! .
	trace_file_number_set_t * unused_traces,
	//!       
	//! .
	std::ostream * verbose_stream,
	//!   .
	reverse_loader_t & loader,
	//!    trace-.
	trace_file_number_set_t & trx_used_traces,
	//!  ,     .
	const oess_1::char_t * image,
	//!   ,     .
	oess_1::uint_t image_len )
	{
		trace_file_number_set_t all_used_traces;

		bool trx_found = true;

		trace_stream_item_t item;
		deserialize_item( image, image_len, item );

		//    --    
		//   .
		oess_1::uint_t total_page_count =
				item.main_file_length() / config.page_size();

		replay_context_t ctx(
				committer,
				total_page_count,
				config.page_size(),
				config.trx_delta_block_size(),
				verbose_stream );

		//    .
		committer.initialize(
				total_page_count, ctx.page_size(), ctx.block_size() );
		do
			{
				if( process_trx( item, ctx ) )
					// -    .  
					//  trace-   .
					std::copy( trx_used_traces.begin(), trx_used_traces.end(),
							std::inserter(
									all_used_traces,
									all_used_traces.begin() ) );

				if( ctx.is_need_continue() )
					{
						//   ,   
						// .
						trx_found = loader.next( image, image_len,
								trx_used_traces );
						if( trx_found )
							deserialize_item( image, image_len, item );
					}
			}
		while( ctx.is_need_continue() && trx_found );

		if( ctx.is_need_continue() &&
				!committer.is_insufficient_trace_info_allowed() )
			{
				ctx.dump();

				OESS_THROW_PHYSIC(
						oess_1::db::storage::err::c_insufficent_trace_info,
						physic_name << ": insufficent amount of information "
								"in trace-stream" )
			}

		committer.finalize();

		calc_used_unused_traces(
				all_traces, all_used_traces,
				used_traces, unused_traces );
	}

//
// do_reverse_replay
//

/*!
	\brief      
		  trace-stream.

	 ,   trace-stream   , 
	    trace_all,   , ..
	  (      
	   ).   ,  
	   trace_last,     
	    ,    commiter.initialize() 
	 0   total_page_count.
*/
static void
do_reverse_replay(
	//!       .
	const std::string & physic_name,
	//!  .
	const config_t & config,
	//!    trace-  trace-stream.
	const trace_file_number_set_t & all_traces,
	//! ,      .
	committer_t & committer,
	//!     trace-,
	//!       .
	trace_file_number_set_t * used_traces,
	//! e    trace-,
	//!     ,   
	//! .
	trace_file_number_set_t * unused_traces,
	//!       
	//! .
	std::ostream * verbose_stream )
	{
		//     trace-.
		//        trace-.
		trace_file_number_set_t trx_used_traces;

		reverse_loader_t loader( physic_name,
				config.trace_file_page_size(),
				all_traces, verbose_stream );

		//       trace-stream,
		//     .
		const oess_1::char_t * image;
		oess_1::uint_t image_len;
		if( loader.next( image, image_len, trx_used_traces ) )
			{
				//   .   
				//   . Trace-,  
				//   ,   
				// .
				trace_file_number_set_t nonactual_traces;
				std::set_difference(
						all_traces.begin(), all_traces.end(),
						trx_used_traces.begin(), trx_used_traces.end(),
						std::inserter(
								nonactual_traces,
								nonactual_traces.begin() ) );

				committer.detect_nonactual_traces( nonactual_traces );

				//      
				//  trace-.
				loader.reduce_traces_set( nonactual_traces );

				//     trace-.
				do_normal_recover(
						physic_name,
						config,
						all_traces,
						committer,
						used_traces,
						unused_traces,
						verbose_stream,
						loader,
						trx_used_traces,
						image,
						image_len );
			}
		else
			//  trace-stream      .
			if( oess_1::db::storage::trace_last == config.trace_mode() )
				{
					//     trace-stream  
					//    .
					do_recover_db_with_damaged_last_trx( committer, config );
				}
			else
				//   , ..    
				// ,     .
				OESS_THROW_PHYSIC(
						oess_1::db::storage::err::c_insufficent_trace_info,
						physic_name << ": unable to recover database without "
						"trx_items at all" )
	}

} /* namespace impl */

//
// std_trace_stream_t
//

std_trace_stream_t::std_trace_stream_t()
	:	m_is_open( false )
	,	m_is_read_only( false )
	{}

std_trace_stream_t::~std_trace_stream_t()
	{}

void
std_trace_stream_t::open(
	const std::string & physic_name,
	const config_t & config,
	bool is_read_only )
	{
		if( m_is_open )
			OESS_THROW_LOGIC(
					oess_1::db::storage::err::c_trace_stream_open,
					"trace-stream already open for: " << m_physic_name
					<< ", new physic-name: " << physic_name )

		//      trace-.
		trace_file_number_set_t all_traces(
				find_all_trace_files( physic_name ) );
		err_code_t rc = impl::check_trace_file_count(
				config.trace_mode(), all_traces );
		if( rc )
			OESS_THROW_PHYSIC( rc.m_code, rc.m_desc )

		impl::initialize_compression( config );

		m_all_traces = all_traces;
		m_physic_name = physic_name;
		m_config = config;
		m_is_open = true;
		m_is_read_only = is_read_only;
	}

void
std_trace_stream_t::create(
	const std::string & physic_name,
	const config_t & config )
	{
		if( m_is_open )
			OESS_THROW_LOGIC(
					oess_1::db::storage::err::c_trace_stream_open,
					"trace-stream already open for: " << m_physic_name
					<< ", new physic-name: " << physic_name )

		//      .
		trace_file_number_set_t all_traces(
				find_all_trace_files( physic_name ) );
		if( all_traces.size() )
			OESS_THROW_PHYSIC(
					oess_1::db::storage::err::c_trace_file_found,
					"some trace-files already exists for: " << physic_name )

		impl::initialize_compression( config );

		//   trace-   .   
		//   trace-stream.
		trace_file_number_t first = 0;
		oess_1::file::file_t first_file;
		first_file.create( create_trace_file_name( physic_name, first ) );

		m_all_traces.insert( first );
		m_physic_name = physic_name;
		m_config = config;
		m_is_open = true;
		m_is_read_only = false;
	}

void
std_trace_stream_t::destroy(
	const std::string & physic_name )
	{
		std::string errors;
		impl::destroy_trace_files(
				physic_name,
				find_all_trace_files( physic_name ),
				errors );

		//         .
		if( find_all_trace_files( physic_name ).size() )
			OESS_THROW_PHYSIC(
					oess_1::db::storage::err::c_trace_file_found,
					"unable to delete all trace files for: "
					<< physic_name << "; all errors: "
					<< errors )
	}

bool
std_trace_stream_t::is_exists(
	const std::string & physic_name )
	{
		//      trace-.
		return find_all_trace_files( physic_name ).size();
	}

void
std_trace_stream_t::append(
	const trace_stream_item_t & item )
	{
		if( !m_is_open )
			OESS_THROW_LOGIC(
					oess_1::db::storage::err::c_trace_stream_not_open,
					"trace-stream not open" )
		if( m_is_read_only )
			OESS_THROW_LOGIC(
					oess_1::db::storage::err::c_read_only_trace_stream,
					"read-only trace-stream for: " << m_physic_name )

		//        
		//    .
//FIX ME:     :
//        
// .        
//    .
		oess_1::io::mem_buf_t buffer;
		impl::serialize_item( item, m_config.trace_compression(), buffer );

		//       trace-.
		if( !m_chain_ostream.get() )
			m_chain_ostream = create_chain_ostream();

		m_chain_ostream->append( buffer.ptr(), buffer.size() );
	}

void
std_trace_stream_t::reverse_replay(
	committer_t & committer,
	trace_file_number_set_t * used_files,
	trace_file_number_set_t * unused_files,
	std::ostream * verbose_stream )
	{
		if( !m_is_open )
			OESS_THROW_LOGIC(
					oess_1::db::storage::err::c_trace_stream_not_open,
					"trace-stream not open" )
		
		impl::do_reverse_replay(
				m_physic_name, m_config, m_all_traces,
				committer, used_files, unused_files,
				verbose_stream );
	}

void
std_trace_stream_t::close()
	{
		m_all_traces.clear();
		m_is_open = false;
	}

std::auto_ptr< impl::chain_ostream_t >
std_trace_stream_t::create_chain_ostream()
	{
		return std::auto_ptr< impl::chain_ostream_t >(
				new impl::chain_ostream_t(
						physic_name(), all_traces(),
						config().trace_file_page_size(),
						config().trace_file_size_limit(),
						config().trace_stream_buffered_pages() ) );
	}

bool
std_trace_stream_t::is_open() const
	{
		return m_is_open;
	}

const std::string &
std_trace_stream_t::physic_name() const
	{
		return m_physic_name;
	}

const config_t &
std_trace_stream_t::config() const
	{
		return m_config;
	}

trace_file_number_set_t &
std_trace_stream_t::all_traces()
	{
		return m_all_traces;
	}

//
// last_trx_only_trace_stream_t
//

last_trx_only_trace_stream_t::last_trx_only_trace_stream_t()
	{}

last_trx_only_trace_stream_t::~last_trx_only_trace_stream_t()
	{}

std::auto_ptr< impl::chain_ostream_t >
last_trx_only_trace_stream_t::create_chain_ostream()
	{
		return std::auto_ptr< impl::chain_ostream_t >(
				new impl::last_trx_only_chain_ostream_t(
						physic_name(), all_traces(),
						config().trace_file_page_size(),
						config().trace_stream_buffered_pages() ) );
	}

//
// pseudo_trace_stream_t
//

pseudo_trace_stream_t::pseudo_trace_stream_t()
	{}

pseudo_trace_stream_t::~pseudo_trace_stream_t()
	{}

void
pseudo_trace_stream_t::open(
	const std::string & physic_name,
	const config_t & config,
	bool is_read_only )
	{}

void
pseudo_trace_stream_t::create(
	const std::string & physic_name,
	const config_t & config )
	{}

void
pseudo_trace_stream_t::destroy(
	const std::string & physic_name )
	{}

bool
pseudo_trace_stream_t::is_exists(
	const std::string & physic_name )
	{
		return true;
	}

void
pseudo_trace_stream_t::append(
	const trace_stream_item_t & item )
	{}

void
pseudo_trace_stream_t::reverse_replay(
	committer_t & committer,
	trace_file_number_set_t * used_files,
	trace_file_number_set_t * unused_files,
	std::ostream * verbose_stream )
	{
		OESS_THROW_PHYSIC(
				oess_1::db::storage::err::c_trace_file_not_found,
				"storage cannot be repaired if trace_mode == trace_no" )
	}

void
pseudo_trace_stream_t::close()
	{}

//
// create_std_trace_stream
//

OESS_1__DB__FUNC(std::auto_ptr< trace_stream_t >)
create_std_trace_stream(
	trace_mode_t trace_mode )
	{
		if( trace_no == trace_mode )
			return std::auto_ptr< trace_stream_t >( 
					new pseudo_trace_stream_t() );

		if( trace_last == trace_mode )
			return std::auto_ptr< trace_stream_t >(
					new last_trx_only_trace_stream_t() );

		if( trace_all != trace_mode )
			OESS_THROW_LOGIC(
					oess_1::db::storage::err::c_invalid_trace_mode,
					trace_mode << ": unknown or unsupported trace mode" )

		return std::auto_ptr< trace_stream_t >(
				new std_trace_stream_t() );
	}

} /* namespace trx */

} /* namespace storage */

} /* namespace db */

} /* namespace oess */

