/*

Copyright (c) 2002-2005, Yauheni Akhotnikau
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

- The name of the author may not be used to endorse or promote products derived
from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.

*/

#include <oess_1/defs/h/ex.hpp>

#include <oess_1/io/h/errno.hpp>
#include <oess_1/io/h/bstring_buf.hpp>

namespace oess_1 {

namespace io {

//
// ibstring_t
//

ibstring_t::ibstring_t(
	const std::string & from ) :
	m_from( from ),
	m_pos( 0 )
{
}

ibstring_t::~ibstring_t()
{
}

size_t
ibstring_t::capacity() const
{
	return m_from.length();
}

size_t
ibstring_t::size() const
{
	return m_from.length();
}

size_t
ibstring_t::pos() const
{
	return m_pos;
}

void
ibstring_t::set_pos( size_t pos )
{
	if( pos <= size() )
		m_pos = pos;
}

bool
ibstring_t::eof() const {
	return ( pos() == size() );
}

const oess_1::char_t *
ibstring_t::in_reserve(
	size_t item_count,
	size_t item_size,
	size_t & item_available )
{
	// ,   
	//  .
	item_available = ( size() - m_pos ) / item_size;
	if( !item_available )
		OESS_THROW_LOGIC( oess_1::io::err::c_eof,
			"size: " << size()
			<< ", pos: " << m_pos
			<< ", item size: " << item_size )

	if( item_available > item_count )
		item_available = item_count;

	size_t bytes_to_read = item_count * item_size;

	return m_from.data() + m_pos;
}

void
ibstring_t::in_shift(
	size_t item_count,
	size_t item_size )
{
	set_pos( m_pos + item_count * item_size );
}

//
// obstring_t
//

obstring_t::obstring_t(
	std::string & to ) :
	m_to( to )
{
}

obstring_t::~obstring_t()
{
}

size_t
obstring_t::len() const
{
	return m_to.length();
}

size_t
obstring_t::capacity() const
{
	return m_to.capacity();
}

size_t
obstring_t::size() const
{
	return m_to.size();
}

size_t
obstring_t::pos() const
{
	return m_to.size();
}

oess_1::char_t *
obstring_t::out_reserve(
	size_t item_count,
	size_t item_size,
	size_t & item_available )
{
	size_t bytes_to_write = item_count * item_size;
	if( bytes_to_write > m_buf.size() ) {
		//     .
		m_buf.destroy();
		m_buf.set( new oess_1::char_t[ bytes_to_write ], bytes_to_write );
	}

	item_available = item_count;

	return m_buf.query();
}

void
obstring_t::out_shift(
	size_t item_count,
	size_t item_size )
{
	m_to.append( m_buf.query(), item_count * item_size );
}

} /* namespace io */

} /* namespace oess_1 */
