/*

Copyright (c) 2002-2005, Yauheni Akhotnikau
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

- The name of the author may not be used to endorse or promote products derived
from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.

*/

#include <oess_1/defs/h/ex.hpp>

#include <oess_1/io/h/errno.hpp>
#include <oess_1/io/h/fixed_mem_buf.hpp>

namespace oess_1 {

namespace io {

//
// ifixed_mem_buf_t
//

ifixed_mem_buf_t::ifixed_mem_buf_t(
	const oess_1::char_t * buf,
	size_t capacity ) :
	m_buf( buf ),
	m_capacity( capacity ),
	m_pos( 0 )
{
}

ifixed_mem_buf_t::~ifixed_mem_buf_t()
{
}

size_t
ifixed_mem_buf_t::capacity() const
{
	return m_capacity;
}

size_t
ifixed_mem_buf_t::size() const
{
	return m_capacity;
}

size_t
ifixed_mem_buf_t::pos() const
{
	return m_pos;
}

void
ifixed_mem_buf_t::set_pos( size_t pos )
{
	if( pos <= m_capacity )
		m_pos = pos;
}

bool
ifixed_mem_buf_t::eof() const
{
	return ( pos() == capacity() );
}

const oess_1::char_t *
ifixed_mem_buf_t::ptr() const
{
	return m_buf;
}

const oess_1::char_t *
ifixed_mem_buf_t::current_ptr() const
{
	return m_buf + m_pos;
}

const oess_1::char_t *
ifixed_mem_buf_t::in_reserve(
	size_t item_count,
	size_t item_size,
	size_t & item_available )
{
	// ,   
	//  .
	item_available = ( m_capacity - m_pos ) / item_size;
	if( !item_available )
		OESS_THROW_LOGIC( oess_1::io::err::c_eof,
			"capacity: " << m_capacity
			<< ", pos: " << m_pos
			<< ", item size: " << item_size )

	if( item_available > item_count )
		item_available = item_count;

	return m_buf + m_pos;
}

void
ifixed_mem_buf_t::in_shift(
	size_t item_count,
	size_t item_size )
{
	set_pos( m_pos + item_count * item_size );
}

//
// ofixed_mem_buf_t
//

ofixed_mem_buf_t::ofixed_mem_buf_t(
	oess_1::char_t * buf,
	size_t capacity,
	size_t initial_size )
:
	m_buf( buf )
,	m_capacity( capacity )
,	m_size( initial_size )
,	m_pos( 0 )
{
}

ofixed_mem_buf_t::~ofixed_mem_buf_t()
{
}

size_t
ofixed_mem_buf_t::capacity() const
{
	return m_capacity;
}

size_t
ofixed_mem_buf_t::size() const
{
	return m_size;
}

size_t
ofixed_mem_buf_t::pos() const
{
	return m_pos;
}

void
ofixed_mem_buf_t::set_pos( size_t pos )
{
	if( pos <= m_size )
		m_pos = pos;
}

const oess_1::char_t *
ofixed_mem_buf_t::ptr() const
{
	return m_buf;
}

const oess_1::char_t *
ofixed_mem_buf_t::current_ptr() const
{
	return m_buf + m_pos;
}

oess_1::char_t *
ofixed_mem_buf_t::out_reserve(
	size_t item_count,
	size_t item_size,
	size_t & item_available )
{
	// ,   
	//  .
	item_available = ( m_capacity - m_pos ) / item_size;
	if( !item_available )
		OESS_THROW_LOGIC( oess_1::io::err::c_eof,
			"capacity: " << m_capacity
			<< ", pos: " << m_pos
			<< ", item size: " << item_size )

	if( item_available > item_count )
		item_available = item_count;

	return m_buf + m_pos;
}

void
ofixed_mem_buf_t::out_shift(
	size_t item_count,
	size_t item_size )
{
	m_pos += item_count * item_size;
	if( m_pos > m_capacity )
		m_pos = m_capacity;
	if( m_pos > m_size )
		m_size = m_pos;
}

} /* namespace io */

} /* namespace oess_1 */
