/*

Copyright (c) 2002-2005, Yauheni Akhotnikau
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

- The name of the author may not be used to endorse or promote products derived
from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.

*/

/*!
	\file oess_1/tlv/h/base.hpp
	\brief      tlv.
*/

#if !defined( _OESS_1__TLV__BASE_HPP_ )
#define _OESS_1__TLV__BASE_HPP_

#include <oess_1/tlv/h/declspec.hpp>

#include <oess_1/io/h/binstream.hpp>

namespace oess_1
{

namespace tlv
{

//! ,       TLV.
/*!
	      Tag 
	 8-, 16-, 32- ,
	, 24- .  ,  Tag
	   - .

	   tlv ,   
	 Tag      
	  defaul_tag_type_t.
*/
typedef oess_1::uint_t	default_tag_type_t;

//
// tag_processor_t
//

//!    Tag   .
/*!
	      Tag.

	   Tag   
	  . ,   Tag 
	 ,  Tag     0  127.
	 ,  Tag     128 
	382.    (b0)  0x80, a  (b1)
	 .  
	Tag   b0  b1.

	   TLV-    
	 Tag,   TLV-  
	   tag_processor_t.
*/
class OESS_1__TLV__TYPE	tag_processor_t
{
	public :
		virtual ~tag_processor_t();

		//!   ,  
		//!    Tag.
		virtual size_t
		tlv_tag_size() const = 0;

		//!  Tag   .
		virtual void
		tlv_unpack_tag(
			//! ,     
			//!    Tag.
			oess_1::io::ibinstream_t & s ) = 0;

		//!  Tag   .
		virtual void
		tlv_pack_tag(
			//! ,     
			//!    Tag.
			oess_1::io::obinstream_t & s ) const = 0;

		//!    Tag.
		virtual default_tag_type_t
		query_tag() const = 0;

		//!    Tag.
		virtual void
		set_tag( const default_tag_type_t & value ) = 0;
};

//
// length_processor_t
//

//!    Length   .
/*!
	      Length.

	   Length   
	  . ,   Length 
	 ,  Length     0  127.
	 ,  Length     128 
	255.    (b0)  0x81, a 
	  (0x80-0xFF). . GSM 11.14 v.8.3.0 1999
	Annex D: Structure of SIM Application Toolkit communications
	(http://www.etsi.org).

	   TLV-    
	 Length,   TLV-  
	   length_processor_t.
*/
class OESS_1__TLV__TYPE	length_processor_t
{
	public :
		virtual ~length_processor_t();

		//!   ,  
		//!    Length.
		virtual size_t
		tlv_length_size(
			//!   TLV.
			size_t value_length ) const = 0;

		//!  Length   .
		virtual void
		tlv_unpack_length(
			//! ,     
			//!    Length.
			oess_1::io::ibinstream_t & s,
			//!    Length
			size_t & value_length ) = 0;

		//!  Length   .
		virtual void
		tlv_pack_length(
			//! ,     
			//!    Length.
			oess_1::io::obinstream_t & s,
			//!   Length.
			size_t value_length ) const = 0;
};

//
// tlv_base_t
//

//!    TLV.
class OESS_1__TLV__TYPE	tlv_base_t
{
	public :
		virtual ~tlv_base_t();

		//!    Tag  TLV.
		/*!
			    TLV,   
			   Tag. , ,
			     TLV.  
			     
			  Tag ("" 
			Tag    ),  
			 .

			\throw std::exception   .
		*/
		virtual default_tag_type_t
		tlv_tag() const = 0;

		//!   ,  
		//!    TLV (  Tag, Length,
		//! Value).
		/*!
			    
			    TLV  , 
			   
			  .
		*/
		virtual size_t
		tlv_size() const = 0;

		//!   Length  Value   .
		/*!
			     ,  
			     Tag.
		*/
		virtual void
		tlv_unpack(
			//! ,     
			//!    Length  Value.
			oess_1::io::ibinstream_t & s,
			//!   Tag,    
			//!    ,    
			//! tlv_unpack.
			default_tag_type_t tag_id ) = 0;

		//!  TLV   .
		/*!
			   (Tag, Length, Value).
		*/
		virtual void
		tlv_pack(
			//! ,     
			//!   TLV.
			oess_1::io::obinstream_t & s ) const = 0;
};

//
// tlv_extractor_t
//

//!    TLV    
//!    (>>).
/*!
	  tlv_base_t::tlv_unpack()   
	   Tag.    ,  
	 TLV      :
\code
some_tag_processor_t tag;
some_tlv_t tlv;
istream >> tag;
tlv.tlv_unpack( istream, tag.query_tag() );
\endcode

	 tlv_extractor_t   
	tag_processor_t  tlv_base_t    
	:
\code
some_tag_processor_t tag;
some_tlv_t tlv;
istream >> tag >> tlv_extractor_t( tag, tlv );
\endcode

*/
class  OESS_1__TLV__TYPE	tlv_extractor_t
{
	public :
		inline
		tlv_extractor_t(
			const tag_processor_t & tag_processor,
			tlv_base_t & tlv )
			:
				m_tag_processor( tag_processor ),
				m_tlv( tlv )
		{
		}

		inline void
		tlv_unpack( oess_1::io::ibinstream_t & s )
		{
			m_tlv.tlv_unpack( s, m_tag_processor.query_tag() );
		}

	private :
		const tag_processor_t &	m_tag_processor;
		tlv_base_t &	m_tlv;
};

//!  TLV   .
/*!
	    ,   
	   Tag.
*/
inline oess_1::io::ibinstream_t &
operator>>(
	//! ,     
	//!    Length  Value.
	oess_1::io::ibinstream_t & s,
	//!    .
	tlv_extractor_t & o )
{
	o.tlv_unpack( s );
	return s;
}

//!   tlv_pack  \a o.
inline oess_1::io::obinstream_t &
operator<<(
	//! ,     
	//!   TLV.
	oess_1::io::obinstream_t & s,
	//!   .
	const tlv_base_t & o )
{
	o.tlv_pack( s );
	return s;
}

} /* namespace tlv */

} /* namespace oess_1 */

#endif

