/*

Copyright (c) 2002-2005, Yauheni Akhotnikau
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

- Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

- The name of the author may not be used to endorse or promote products derived
from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
OF SUCH DAMAGE.

*/

/*!
	\file oess_1/tlv/h/compound.hpp
	\brief  %oess_1::tlv::coumpound_tlv_t.
*/

#if !defined( _OESS_1__TLV__COMPOUND_HPP_ )
#define _OESS_1__TLV__COMPOUND_HPP_

#include <oess_1/tlv/h/declspec.hpp>

#include <map>
#include <string>
#include <functional>
#include <algorithm>

#include <cpp_util_2/h/nocopy.hpp>

#include <oess_1/defs/h/ex.hpp>

#include <oess_1/io/h/subbinstream.hpp>

#include <oess_1/tlv/h/errno.hpp>
#include <oess_1/tlv/h/base.hpp>
#include <oess_1/tlv/h/default_processors.hpp>

namespace oess_1
{

namespace tlv
{

//
// tlv_add_caller_t
//

//!     
//!  TLV   TLV.
/*!
	\since v.1.0.9

	  compound_tlv_t   :
	  TLV     
	TLV  compound_tlv_t::tlv_add().   
	TLV    TLV,    
	tlv_add  -  .    
	.

	 tlv_add_caller_t     :
	-    TLV,   
	 compound_tlv_t::tlv_add(),   
	tlv_add_caller_t;
	-       
	   compound_tlv_t:

\code
typedef oess_1::tlv::compound_tlv_t<tag_type_t > compound_t;
class	my_compound_tlv_t :
	public compound_t
{
	private :
		oess_1::tlv::tagged_scalar_wrapper_t<
				0x02, string_tlv_t >
			m_member_1;
		const tlv_add_t	m_member_1_inserter;

		oess_1::tlv::tagged_scalar_wrapper_t<
				0x05, oess_1::tlv::scalar_tlv_t<
					tag_type_t, oess_1::ushort_t > >
			m_member_2;
		const tlv_add_t	m_member_2_inserter;

	public :
		my_compound_tlv_t(
			const std::string & member_1,
			oess_1::ushort_t member_2 )
			:
				compound_t( 0x01 ),
				m_member_1( member_1 ),
				m_member_1_inserter( tlv_self(), m_member_1 ),

				m_member_2( member_2 ),
				m_member_2_inserter( tlv_self(), m_member_2 ),
		{
		}

		my_compound_tlv_t()
			:
				compound_t( 0x01 ),
				m_member_1_inserter( tlv_self(), m_member_1 ),
				m_member_2_inserter( tlv_self(), m_member_2 )
		{
		}
};
\endcode
*/
template< class Compound >
class	tlv_add_caller_t
{
	public :
		/*!
			   tlv_add()   owner.
		*/
		tlv_add_caller_t(
			//!  TLV.
			Compound & owner,
			//!  TLV.
			tlv_base_t & nested )
		{
			owner.tlv_add( nested );
		}
};

//
// compound_tlv_t
//

//! TLV    .
/*!
	 compound_tlv_t   .

\par  (. oess_1::tlv::basic_string_tlv_t)
\code
typedef oess_1::tlv::compound_tlv_t<oess_1::ushort_t > compound_t;
class	my_compound_tlv_t :
	public compound_t
{
	private :
		oess_1::tlv::basic_string_tlv_t<
				oess_1::ushort, oess_1::char_t >
			m_member_1;
		oess_1::tlv::basic_string_tlv_t<
				oess_1::ushort, oess_1::char_t >
			m_member_2;
	public :
		my_compound_tlv_t() :
			compound_t( 0x01 ),
			m_member_1( 0x02 ),
			m_member_2( 0x03 )
		{
			tlv_add( 0x02, m_member_1 );
			tlv_add( 0x03, m_member_2 );
		}
};
\endcode
*/
template<
	class Tag_type,
	class Tag_processor = default_tag_processor_t< Tag_type >,
	class Length_processor = default_length_processor_t< Tag_type > >
class	compound_tlv_t :
	public tlv_base_t,
	private cpp_util_2::nocopy_t
{
	public :
		//!   compound_tlv_t< Tag_type,
		//! Tag_processor, Length_processor >.
		typedef compound_tlv_t< Tag_type,
				Tag_processor, Length_processor >
			self_t;

		//!   %tlv_add_caller_t 
		//!   TLV.
		/*! \since v.1.0.9 */
		typedef tlv_add_caller_t< self_t >	tlv_add_t;

	private :
		//!    TLV.
		typedef std::map< default_tag_type_t, tlv_base_t * >
			tlv_map_t;

		//!   std::for_each.
		/*!
			    TLV.
		*/
		class	length_calc_t :
			public std::unary_function<
				const typename tlv_map_t::value_type &, void >
		{
			private :
				size_t &	m_length;
			public :
				length_calc_t( size_t & length )
					:
						m_length( length )
				{
				}

				result_type
				operator()( argument_type a )
				{
					m_length += a.second->tlv_size();
				}
		};

		//!   std::for_each.
		/*!
			   TLV.
		*/
		class	packer_t :
			public std::unary_function<
				const typename tlv_map_t::value_type &, void >
		{
			private :
				oess_1::io::obinstream_t &	m_stream;
			public :
				packer_t( oess_1::io::obinstream_t & stream )
					:
						m_stream( stream )
				{
				}

				result_type
				operator()( argument_type a )
				{
					a.second->tlv_pack( m_stream );
				}
		};

		//!   Tag.
		Tag_processor	m_tag;

		//!   Length.
		Length_processor	m_length;

		//!  TLV.
		tlv_map_t	m_tlvs;

		//!   TLV  .
		/*!
			,        Tag.
		*/
		void
		try_unpack_tlv(
			//!  Tag,   .
			default_tag_type_t tag,
			//! ,    .
			oess_1::io::ibinstream_t & s )
		{
			typename tlv_map_t::iterator it = m_tlvs.find( tag );
			if( it != m_tlvs.end() )
				//  .
				it->second->tlv_unpack( s, tag );
			else
				//   .
				tlv_on_unexpected_tag( tag, s );
		}


	protected :
		//!    TLV,  
		//!    .
		/*!
			     
			oess_1::io::physic_ex_t  
			oess_1::tlv::err::c_unexpected_tag.
		*/
		virtual void
		tlv_on_unexpected_tag(
			//!  Tag,   .
			default_tag_type_t tag,
			//! ,    .
			/*!
				    
				      TLV.
			*/
			oess_1::io::ibinstream_t & s )
		{
			OESS_THROW_PHYSIC( oess_1::tlv::err::c_unexpected_tag,
				"compound tag: " << m_tag.query_tag()
				<< ", child tag found: " << tag )
		}

	public :
		compound_tlv_t( Tag_type id )
			:
				m_tag( id )
		{
		}

		virtual ~compound_tlv_t()
		{
		}

		//!   Tag.
		virtual default_tag_type_t
		tlv_tag() const
		{
			return m_tag.query_tag();
		}

		//!   TLV.
		/*!
			  compound_tlv_t    TLV
			Tag,  tlv.tlv_tag(),    .
		*/
		void
		tlv_add(
			//!  TLV.
			/*!
				     ,
				     compound_tlv_t.
			*/
			tlv_base_t & tlv )
		{
			m_tlvs[ tlv.tlv_tag() ] = &tlv;
		}

		//!   TLV    Tag.
		/*!
			  compound_tlv_t    TLV  
			Tag,    .

			    , 
			  tlv   TLV 
			 Tag (.. tag != tlv.tlv_tag()).

			  ,   tlv  
			   TLV,   
			   Tag.
		*/
		void
		tlv_add(
			//!  Tag   TLV.
			default_tag_type_t tag,
			//!  TLV.
			/*!
				     ,
				     compound_tlv_t.
			*/
			tlv_base_t & tlv )
		{
			m_tlvs[ tag ] = &tlv;
		}

		//!   TLV    Tag.
		/*!
			  TLV   .
		*/
		void
		tlv_remove(
			//!   TLV.
			default_tag_type_t tag )
		{
			m_tlvs.erase( tag );
		}

		//!  ,  
		//!    TLV (  Tag, Length,
		//! Value).
		virtual size_t
		tlv_size() const
		{
			size_t length = 0;
			std::for_each( m_tlvs.begin(), m_tlvs.end(),
				length_calc_t( length ) );

			length += m_length.tlv_length_size( length );
			length += m_tag.tlv_tag_size();

			return length;
		}

		//!   TLV  .
		virtual void
		tlv_unpack(
			//! ,     
			//!    Length  Value.
			oess_1::io::ibinstream_t & s,
			//!    .
			default_tag_type_t )
		{
			size_t length;
			m_length.tlv_unpack_length( s, length );

			//    TLV.
			oess_1::io::isubbinstream_t tmp_istream( s, length );
			while( !tmp_istream.eof() )
			{
				Tag_processor tag;
				tag.tlv_unpack_tag( tmp_istream );

				try_unpack_tlv( tag.query_tag(), tmp_istream );
			}
		}

		//!  TLV   .
		virtual void
		tlv_pack(
			//! ,     
			//!   TLV.
			oess_1::io::obinstream_t & s ) const
		{
			//    .
			size_t length = 0;
			std::for_each( m_tlvs.begin(), m_tlvs.end(),
				length_calc_t( length ) );

			m_tag.tlv_pack_tag( s );
			m_length.tlv_pack_length( s, length );
			if( length )
			{
				packer_t packer( s );
				std::for_each( m_tlvs.begin(), m_tlvs.end(), packer );
			}
		}

		//!     .
		/*! \since v.1.0.9

			   
			tlv_add_caller_t. */
		self_t &
		tlv_self()
		{
			return *this;
		}
};

} /* namespace tlv */

} /* namespace oess_1 */

#endif

